"""
Sub-tasks that serve as children to other tasks.

.. note::

    The Child agent will be formalized in an upcoming release, until then these classes
    remain relatively undocumented as their design will likely change.

"""

from collections import OrderedDict as odict
from collections import deque

import autopilot.transform
from autopilot import prefs
from autopilot.hardware.gpio import Digital_Out
from autopilot.hardware.usb import Wheel
from autopilot.hardware import cameras
from autopilot.core.networking import Net_Node
from autopilot.transform import transforms
from itertools import cycle
from queue import Empty, LifoQueue
import threading
import logging
from time import sleep

class Wheel_Child(object):
    STAGE_NAMES = ['collect']

    PARAMS = odict()
    PARAMS['fs'] = {'tag': 'Velocity Reporting Rate (Hz)',
                    'type': 'int'}
    PARAMS['thresh'] = {'tag': 'Distance Threshold',
                        'type': 'int'}

    HARDWARE = {
        "OUTPUT": Digital_Out,
        "WHEEL":  Wheel
    }



    def __init__(self, stage_block=None, fs=10, thresh=100, **kwargs):
        self.fs = fs
        self.thresh = thresh

        self.hardware = {}
        self.hardware['OUTPUT'] = Digital_Out(prefs.HARDWARE['OUTPUT'])
        self.hardware['WHEEL'] = Wheel(digi_out = self.hardware['OUTPUT'],
                                       fs       = self.fs,
                                       thresh   = self.thresh,
                                       mode     = "steady")
        self.stages = cycle([self.noop])
        self.stage_block = stage_block

    def noop(self):
        # just fitting in with the task structure.
        self.stage_block.clear()
        return {}

    def end(self):
        self.hardware['WHEEL'].release()
        self.stage_block.set()


class Video_Child(object):
    PARAMS = odict()
    PARAMS['cams'] = {'tag': 'Dictionary of camera params, or list of dicts',
                      'type': ('dict', 'list')}

    def __init__(self, cams=None, stage_block = None, start_now=True, **kwargs):
        """
        Args:
            cams (dict, list): Should be a dictionary of camera parameters or a list of dicts. Dicts should have, at least::

                {
                    'type': 'string_of_camera_class',
                    'name': 'name_of_camera_in_task',
                    'param1': 'first_param'
                }
        """

        if cams is None:
            Exception('Need to give us a cams dictionary!')

        self.cams = {}

        self.start_now = start_now


        if isinstance(cams, dict):

            try:
                cam_class = getattr(cameras, cams['type'])
                self.cams[cams['name']] = cam_class(**cams)
                # if start:
                #     self.cams[cams['name']].capture()
            except AttributeError:
                AttributeError("Camera type {} not found!".format(cams['type']))

        elif isinstance(cams, list):
            for cam in cams:
                try:
                    cam_class = getattr(cameras, cam['type'])
                    self.cams[cam['name']] = cam_class(**cam)
                    # if start:
                    #     self.cams[cam['name']].capture()
                except AttributeError:
                    AttributeError("Camera type {} not found!".format(cam['type']))

        self.stages = cycle([self.noop])
        self.stage_block = stage_block


        if self.start_now:
            self.start()
        # self.thread = threading.Thread(target=self._stream)
        # self.thread.daemon = True
        # self.thread.start()

    def start(self):
        for cam in self.cams.values():
            cam.capture()

    def stop(self):
        for cam_name, cam in self.cams.items():
            try:
                cam.release()
            except Exception as e:
                Warning('Couldnt release camera {},\n{}'.format(cam_name, e))



    def _stream(self):
        self.node = Net_Node(
            "T_CHILD",
            upstream=prefs.NAME,
            port=prefs.MSGPORT,
            listens = {},
            instance=True
        )

        while True:
            for name, cam in self.cams.items():
                try:
                    frame, timestamp = cam.q.get_nowait()
                    self.node.send(key='CONTINUOUS',
                                   value={cam.name:frame,
                                          'timestamp':timestamp},
                                   repeat=False,
                                   flags={'MINPRINT':True})
                except Empty:
                    pass



    def noop(self):
        # just fitting in with the task structure.
        self.stage_block.clear()
        return {}

    # def start(self):
    #     for cam in self.cams.values():
    #         cam.capture()
    #
    # def stop(self):
    #     for cam in self.cams.values():
    #         cam.release()

class Transformer(object):

    def __init__(self, transform,
                 operation: str ="trigger",
                 return_id = 'T',
                 return_key = None,
                 stage_block = None,
                 value_subset=None,
                 **kwargs):
        """

        Args:
            transform:
            operation (str): either

                * "trigger", where the last transform is a :class:`~autopilot.transform.transforms.Condition`
                and a trigger is returned to sender only when the return value of the transformation changes, or
                * "stream", where each result of the transformation is returned to sender

            return_id:
            stage_block:
            value_subset (str): Optional - subset a value from from a dict/list sent to :meth:`.l_process`
            **kwargs:
        """
        assert operation in ('trigger', 'stream', 'debug')
        self.operation = operation
        self._last_result = None

        if return_key is None:
            self.return_key = self.operation.upper()
        else:
            self.return_key = return_key

        self.return_id = return_id
        self.stage_block = stage_block
        self.stages = cycle([self.noop])
        # self.input_q = LifoQueue()
        self.input_q = deque(maxlen=1)
        self.value_subset = value_subset

        self.logger = logging.getLogger('main')

        self.process_thread = threading.Thread(target=self._process, args=(transform,))
        self.process_thread.daemon = True
        self.process_thread.start()

    def noop(self):
        # just fitting in with the task structure.
        self.stage_block.clear()
        return {}



    def _process(self, transform):

        self.transform = autopilot.transform.make_transform(transform)

        self.node = Net_Node(
            f"{prefs.NAME}_TRANSFORMER",
            upstream=prefs.NAME,
            port=prefs.MSGPORT,
            listens = {
                'CONTINUOUS': self.l_process
            },
            instance=False
        )

        self.node.send(self.return_id, 'STATE', value='READY')

        while True:
            try:
                # value = self.input_q.get_nowait()
                value = self.input_q.popleft()
            # except Empty:
            except IndexError:
                sleep(0.001)
                continue
            result = self.transform.process(value)

            self.node.logger.debug(f'Processed frame, result: {result}')


            if self.operation == "trigger":
                if result != self._last_result:
                    self.node.send(self.return_id, self.return_key, result)
                    self._last_result = result

            elif self.operation == 'stream':
                # FIXME: Another key that's not TRIGGER
                self.node.send(self.return_id, self.return_key, result)

            elif self.operation == 'debug':
                pass


    def l_process(self, value):
        # get array out of value

        # FIXME hack for dlc
        self.node.logger.debug('Received and queued processing!')
        # self.input_q.put_nowait(value['MAIN'])
        if self.value_subset:
            value = value[self.value_subset]
        self.input_q.append(value)













import numpy as np

from autopilot.transform.transforms import Transform


class Condition(Transform):
    """
    Compare the input against some condition
    """


    def __init__(self, minimum=None, maximum=None, elementwise=False, *args, **kwargs):
        """

        Args:
            minimum:
            maximum:
            elementwise (bool): if False, return True only if *all* values are within range. otherwise return bool for each tested value
            *args:
            **kwargs:
        """

        if minimum is None and maximum is None:
            raise ValueError("need either a maximum or minimum!")

        super(Condition, self).__init__(*args, **kwargs)

        self._minimum = None
        self._maximum = None
        self._shape = None
        self.elementwise = elementwise

        if minimum is not None:
            self.minimum = minimum

        if maximum is not None:
            self._maximum = maximum


    def process(self, input):

        if self.minimum is not None:
            is_greater = np.greater(input, self.minimum)
            if self.maximum is None:
                combined = is_greater

        if self.maximum is not None:
            is_lesser = np.less(input, self.maximum)
            if self.minimum is None:
                combined = is_lesser

        if self.minimum is not None and self.maximum is not None:
            combined = np.logical_and(is_greater, is_lesser)

        if not self.elementwise:
            combined = np.all(combined)

        return combined






    @property
    def minimum(self) -> [np.ndarray, float]:
        return self._minimum

    @minimum.setter
    def minimum(self, minimum: [np.ndarray, float]):

        if isinstance(minimum, list):
            minimum = np.array(minimum)

        if isinstance(minimum, float) or isinstance(minimum, int):
            shape = (1,)
        elif isinstance(minimum, np.ndarray):
            shape = minimum.shape
        else:
            raise ValueError('minimum must be a float or ndarray')

        if self._shape:
            if shape != self._shape:
                raise ValueError('cant change shape!')


        self._shape = shape
        self._minimum = minimum

    @property
    def maximum(self) -> [np.ndarray, float]:
        return self._maximum

    @maximum.setter
    def maximum(self, maximum: [np.ndarray, float]):
        if isinstance(maximum, list):
            minimum = np.array(maximum)

        if isinstance(maximum, float) or isinstance(maximum, int):
            shape = (1,)
        elif isinstance(maximum, np.ndarray):
            shape = maximum.shape
        else:
            raise ValueError('maximum must be a float or ndarray')

        if self._shape:
            if shape != self._shape:
                raise ValueError('cant change shape!')

        self._shape = shape
        self._maximum = maximum

    @property
    def format_in(self) -> dict:
        if self._shape == (1,):
            ret = {
                'type': float,

            }
        else:
            ret = {
                'type': np.ndarray
            }

        ret['shape'] = self._shape

        return ret


    @property
    def format_out(self) -> dict:
        if self._shape == (1,):
            ret = {
                'type': bool,
            }
        else:
            ret = {
                'type': np.ndarray
            }

        if self.elementwise:
            ret['shape'] = self._shape
        else:
            ret['type'] = bool
            ret['shape'] = (1,)

        return ret


from autopilot import prefs

from autopilot.stim.managers import Stim_Manager, Proportional, init_manager

if prefs.AGENT == "pilot":
    if 'AUDIO' in prefs.CONFIG:
        from autopilot.stim.sound import sounds



import threading
from subprocess import Popen, PIPE
import sys
import os
import csv
from skvideo import io
from skvideo.utils import vshape
import numpy as np
import base64
from datetime import datetime
import multiprocessing as mp
from tqdm.auto import tqdm


import time
import traceback
import blosc
import warnings
import subprocess
import logging
from ctypes import c_char_p




# import the Queue class from Python 3
if sys.version_info >= (3, 0):
    from queue import Queue, Empty, Full

# otherwise, import the Queue class for Python 2.7
else:
    from Queue import Queue, Empty, Full

try:
    import PySpin
    PYSPIN = True
    PYSPIN_SYSTEM = None
except:
    PYSPIN = False

try:
    import cv2
    OPENCV = True
except:
    OPENCV = False

from autopilot import prefs
from autopilot.hardware import Hardware

OPENCV_LAST_INIT_TIME = mp.Value('d', 0.0)
"""
Time the last OpenCV camera was initialized (seconds, from time.time()).

v4l2 has an extraordinarily obnoxious ...feature -- 
if you try to initialize two cameras at ~the same time,
you will get a neverending stream of informative error messages: ``VIDIOC_QBUF: Invalid argument``

The workaround seems to be relatively simple, we just wait ~2 seconds if another camera was just initialized.
"""
LAST_INIT_LOCK = mp.Lock()

class Camera(Hardware):
    """
    Metaclass for Camera objects. Should not be instantiated on its own.

    Arguments:
        fps (int): Framerate of video capture
        timed (bool, int, float): If False (default), camera captures indefinitely. If int or float, captures for this many seconds
        **kwargs: Arguments to :meth:`~Camera.stream`, :meth:`~Camera.write`, and :meth:`~Camera.queue` can be passed as dictionaries, eg.::

            stream={'to':'T', 'ip':'localhost'}

    When the camera is instantiated and :meth:`~.Camera.capture` is called,
    the class uses a series of methods that should be overwritten in subclasses.
    Further details for each can be found in the relevant method documentation.

    It is highly recommended to instantiate Cameras with a :attr:`.Hardware.name`,
    as it is used in :attr:`.output_filename` and to identify the network stream

    Three methods are required to be overwritten by all subclasses:

        * :meth:`~.Camera.init_cam` - **required** - used by :attr:`~.Camera.cam`, instantiating the camera object so that it can be queried and configured
        * :meth:`~.Camera._grab` - **required** - grab a frame from the :attr:`~.Camera.cam`
        * :meth:`~.Camera._timestamp` - **required** - get a timestamp for the frame

    The other methods are optional and depend on the particular camera:

        * :meth:`~.Camera.capture_init` - *optional* - any required routine to prepare the camera after it is instantiated but before it begins to capture
        * :meth:`~.Camera._process` - *optional* - the wrapper around a full acquisition cycle, including streaming, writing, and queueing frames
        * :meth:`~.Camera._write_frame` - *optional* - how to write an individual frame to disk
        * :meth:`~.Camera._write_deinit` - *optional* - any required routine to finish writing to disk after acquisition
        * :meth:`~.Camera.capture_deinit` - *optional* - any required routine to stop acquisition but not release the camera instance.

    Attributes:
        frame (tuple): The current captured frame as a tuple (timestamp, frame).
        shape (tuple): Shape of captured frames (height, width, channels)
        blosc (bool): If True (default), use blosc compression when
        cam: The object used to interact with the camera
        fps (int): Framerate of video capture
        timed (bool, int, float): If False (default), camera captures indefinitely. If int or float, captures for this many seconds
        q (Queue): Queue that allows frames to be pulled by other objects
        queue_size (int): How many frames should be buffered in the queue.
        initialized (threading.Event): Called in :meth:`~.init_cam` to indicate the camera has been initialized
        stopping (threading.Event): Called to signal that capturing should stop. when set, ends the threaded capture loop
        capturing (threading.Event): Set when camera is actively capturing
        streaming (threading.Event): Set to indicate that the camera is streaming data over the network
        writing (threading.Event): Set to indicate that the camera is writing video locally
        queueing (threading.Event): Indicates whether frames are being put into :attr:`~.Camera.q`
        indicating (threading.Event): Set to indicate that capture progress is being indicated in stdout by :class:`~tqdm.tqdm`


    """
    input = True #: test documenting input
    type = "CAMERA" #: (str): what are we anyway?
    trigger = False

    def __init__(self, fps=None, timed=False, crop=None, **kwargs):
        """

        Args:
            fps:
            timed:
            crop (tuple): (x, y of top left corner, width, height)
            **kwargs:
        """
        super(Camera, self).__init__(**kwargs)

        # internal attributes
        self._cam = None #: camera subobject test
        self._output_filename = None
        self._capture_thread = None
        self._writer = None
        self._write_q = None
        self._stream_q = None
        self._indicator = None

        self.frame = None
        self.shape = None
        self.frame_n = 0
        self.crop = crop

        self.blosc = True

        #self.fps = fps
        self.timed = timed

        self.q = None
        self.queue_size = None

        self.initialized = threading.Event()
        self.initialized.clear()

        # event to end acquisition
        self.stopping = threading.Event()
        self.stopping.clear()

        self.capturing = threading.Event()
        self.capturing.clear()

        self.streaming = threading.Event()
        self.streaming.clear()

        self.writing = threading.Event()
        self.writing.clear()

        self.queueing = threading.Event()
        self.queueing.clear()

        self.indicating = threading.Event()
        self.indicating.clear()

        # initialize args passed by kwargs
        if 'stream' in kwargs.keys():
            self.stream(**kwargs['stream'])

        if 'write' in kwargs.keys():
            self.write(**kwargs['write'])

        if 'queue' in kwargs.keys():
            self.queue(**kwargs['queue'])

    def capture(self, timed = None):
        """
        Spawn a thread to begin capturing.


        Args:
            timed (None, int, float): if None, record according to :attr:`.timed` (default). If numeric, record for ``timed`` seconds.
        """


        if self.capturing.is_set():
            self.logger.warning("Already Capturing!")
            return

        if timed:
            self.timed = timed

        self.frame_n = 0

        self._capture_thread = threading.Thread(target=self._capture)
        self._capture_thread.setDaemon(True)
        self._capture_thread.start()

    def _capture(self):
        """
        Threaded capture method started by :meth:`.capture`.

        Captures until :attr:`.stopping` is set.

        Calls capture methods, in order:

        * :meth:`~.Camera.capture_init` - any required routine to prepare the camera after it is instantiated but before it begins to capture
        * :meth:`~.Camera._process`  - the wrapper around a full acquisition cycle, including streaming, writing, and queueing frames
        * :meth:`~.Camera._grab`  - grab a frame from the :attr:`~.Camera.cam`
        * :meth:`~.Camera._timestamp`  - get a timestamp for the frame
        * :meth:`~.Camera._write_frame`  - how to write an individual frame to disk
        * :meth:`~.Camera._write_deinit` - any required routine to finish writing to disk after acquisition
        * :meth:`~.Camera.capture_deinit` - any required routine to stop acquisition but not release the camera instance.



        Returns:

        """

        self.capturing.set()
        self.stopping.clear()

        self.capture_init()

        if self.streaming.is_set():
            self.node.send(key='STATE', value='CAPTURING')

        try:
            self._process()
            if isinstance(self.timed, int) or isinstance(self.timed, float):
                if self.timed > 0:
                    start_time = time.time()
                    end_time = start_time + self.timed

            while not self.stopping.is_set():
                self._process()

                if self.timed:
                    if time.time() >= end_time:
                        self.stopping.set()

                self.frame_n += 1

        finally:
            self.logger.info('Capture Ending')

            try:
                if self.streaming.is_set():
                    self.node.send(key='STATE', value='STOPPING')
                    self._stream_q.put('END')
            except Exception as e:
                self.logger.exception('Failed to end stream, error message: {}'.format(e))

            try:
                if self.writing.is_set():
                    self._write_deinit()

            except Exception as e:
                self.logger.exception('Failed to end writer, error message: {}'.format(e))

            if self.indicating.is_set():
                try:
                    self._indicator.close()
                except:
                    pass

            self.capturing.clear()
            self.capture_deinit()
            #self.release()
            #self.logger.info('Camera Released')

    def _process(self):
        """
        A full frame capture cycle.

        :meth:`~Camera._grab`s the :attr:`.frame`, then handles streaming, writing, queueing, and indicating
        according to :meth:`~Camera.stream`, :meth:`~Camera.write`, :meth:`~Camera.queue`, and :attr:`~Camera.indicating`, respectively.

        """

        try:
            self.frame = self._grab()
        except Exception as e:
            self.logger.exception(e)

        if self.streaming.is_set():
            self._stream_q.put_nowait({'timestamp': self.frame[0],
                                       self.name  : self.frame[1]})

        if self.writing.is_set():
            self._write_frame()

        if self.queueing.is_set():
            self.q.put_nowait(self.frame)

        if self.indicating.is_set():
            if not self._indicator:
                self._indicator = tqdm()
            self._indicator.update()

    def stream(self, to='T', ip=None, port=None, min_size=5, **kwargs):
        """
        Enable streaming frames on capture.

        Spawns a :class:`~.networking.Net_Node` with :meth:`.Hardware.init_networking`,
        and creates a streaming queue with :meth:`.Net_Node.get_stream` according to args.

        Sets :attr:`.Camera.streaming`

        Args:
            to (str): ID of the recipient. Default 'T' for Terminal.
            ip (str): IP of recipient. If None (default), 'localhost'. If None and ``to`` is 'T', ``prefs.TERMINALIP``
            port (int, str): Port of recipient socket. If None (default), ``prefs.MSGPORT``. If None and ``to`` is 'T', ``prefs.TERMINALPORT``.
            **kwargs: passed to :meth:`.Hardware.init_networking` and thus to :class:`.Net_Node`

        """


        if to=='T':
            if not ip:
                ip = prefs.TERMINALIP
            if not port:
                port = prefs.TERMINALPORT

        else:

            if not ip:
                self.logger.warning('ip not passed, using localhost as default')
                ip = 'localhost'
            if not port:
                self.logger.warning('port not passed, using prefs.MSGPORT')
                port = prefs.MSGPORT


        self.listens = {
            'START': self.l_start,
            'STOP': self.l_stop
        }

        self.init_networking(listens=self.listens, **kwargs)

        if hasattr(prefs, 'SUBJECT'):
            subject = prefs.SUBJECT
        else:
            self.logger.warning('nothing found for prefs.SUBJECT, probably running outside of task context')
            subject = None

        self._stream_q = self.node.get_stream(
            'stream', 'CONTINUOUS', upstream=to,
            ip=ip, port=port, subject=subject,
            min_size=min_size
        )

        self.streaming.set()

    def l_start(self, val):
        """
        Begin capturing by calling :meth:`Camera.capture`

        Args:
            val: unused
        """
        self.capture()

    def l_stop(self, val):
        """
        Stop capture by calling :meth:`Camera.release`

        Args:
            val: unused
        """
        self.release()



    def write(self, output_filename = None, timestamps=True, blosc=True):
        """
        Enable writing frames locally on capture

        Spawns a :class:`.Video_Writer` to encode video, sets :attr:`.writing`

        Args:
            output_filename (str): path and filename of the output video. extension should be ``.mp4``,
                as videos are encoded with libx264 by default.
            timestamps (bool): if True, (timestamp, frame) tuples will be put in the :attr:`._write_q`.
                if False, timestamps will be generated by :class:`.Video_Writer` (not recommended at all).
            blosc (bool): if true, compress frames with :func:`blosc.pack_array` before putting in :attr:`._write_q`.
        """
        if not output_filename:
            output_filename = self.output_filename
        else:
            self._output_filename = output_filename

        self.blosc = blosc
        self._write_q = mp.Queue()
        self.writer = Video_Writer(self._write_q, output_filename, self.fps, timestamps=timestamps, blosc=blosc)
        self.writer.start()
        self.writing.set()
        self.logger.info('Writing initialized, writing to {}'.format(output_filename))

    def _write_frame(self):
        """
        Put :attr:`.frame` into the :attr:`._write_q`, optionally compressing it with :func:`blosc.pack_array`
        """
        try:
            if self.blosc:
                self._write_q.put_nowait((self.frame[0], blosc.pack_array(self.frame[1])))
            else:
                self._write_q.put_nowait(self.frame)
        except Full:
            self.logger.exception('Frame {} could not be written, queue full'.format(self.frame_n))



    def _write_deinit(self):
        """
        End the :class:`.Video_Writer`.

        Blocks until the :attr:`._write_q` is empty, holding the release of the object.
        """
        self._write_q.put_nowait('END')
        checked_empty = False
        while not self._write_q.empty():
            if not checked_empty:
                self.logger.warning(
                    'Writer still has ~{} frames, waiting on it to finish'.format(self._write_q.qsize()))
                checked_empty = True
            time.sleep(0.1)
        self.logger.info('Writer finished, closing')

    def queue(self, queue_size = 128):
        """
        Enable stashing frames in a queue for a local consumer.

        Other objects can get frames as they are acquired from :attr:`.q`

        Args:
            queue_size (int): max number of frames that can be held in :attr:`~Camera.q`
        """
        self.queue_size = queue_size
        self.q = Queue(maxsize=self.queue_size)
        self.queueing.set()
        self.logger.info('Queueing initialized, queue size {}'.format(queue_size))


    @property
    def cam(self):
        """
        Camera object.

        If :attr:`._cam` hasn't been initialized yet, use :meth:`.init_cam` to do so

        Returns:
            Camera object, different for each camera.
        """
        if not self._cam:
            self._cam = self.init_cam()
        return self._cam

    @property
    def output_filename(self):
        """
        Filename given to video writer.

        If explicitly set, returns as expected.

        If None, or path already exists while the camera isn't capturing,
        a new filename is generated in the user directory.

        Returns:
            (str) :attr:`._output_filename`

        """
        # TODO: choose output directory

        new = False
        if self._output_filename is None:
            new = True
        elif os.path.exists(self._output_filename) and not self.capturing.is_set():
            new = True

        if new:
            user_dir = os.path.expanduser('~')
            self._output_filename = os.path.join(user_dir, "capture_{}_{}.mp4".format(self.name,
                                                                                            datetime.now().strftime(
                                                                                                "%y%m%d-%H%M%S")))

        return self._output_filename

    @output_filename.setter
    def output_filename(self, output_filename):
        self._output_filename = output_filename

    def _grab(self):
        """
        Capture a frame and timestamp.

        Method must be overridden by subclass

        Returns:
            (str, :class:`numpy.ndarray`) Tuple of isoformatted (str) or numeric timestamp returned by :meth:`._timestamp`,
                and captured frame
        """
        raise Exception("internal _grab method must be overwritten by camera subclass!!")

    def _timestamp(self, frame=None):
        """
        Generate a timestamp for each :meth:`~Camera._grab`

        Must be overridden by subclass

        Args:
            frame: If needed by camera subclass, pass the frame or image object to get timestamp

        Returns:
            (str, int, float) Either an isoformatted (str) or numeric timestamp

        """
        raise Exception("internal _timestamp method must be overwritten by camera subclass!!")


    def init_cam(self):
        """
        Method to initialize camera object

        Must be overridden by camera subclass

        Returns:
            camera object

        """
        raise Exception('init_cam must be overwritten by camera subclass!!')

    def capture_init(self):
        """
        Optional: Prepare :attr:`.cam` after initialization, but before capture

        Returns:
            None
        """
        pass

    def capture_deinit(self):
        """
        Optional: Return :attr:`.cam` to an idle state after capturing, but before releasing

        Returns:
            None
        """

    def stop(self):
        """
        Stop capture by setting  :attr:`.stopping`
        """
        self.stopping.set()

    def release(self):
        """
        Release resources held by Camera.

        Must be overridden by subclass.

        Does not raise exception in case some general camera release logic should be put here...
        """

        pass
        # raise Exception('release must be overwritten by camera subclass!!')






class Camera_CV(Camera):
    def __init__(self, camera_idx = 0, **kwargs):
        """
        Capture Video from a webcam with OpenCV

        By default, OpenCV will select a suitable backend for the indicated camera. Some backends have difficulty
        operating multiple cameras at once, so the performance of this class will be variable depending on camera
        type.

        Args:
            camera_idx (int): The index of the desired camera
            **kwargs: Passed to the :class:`.Camera` metaclass.

        Attributes:
            camera_idx (int): The index of the desired camera
            last_opencv_init (float): See :data:`~cameras.OPENCV_LAST_INIT_TIME`
            last_init_lock (:class:`threading.Lock`): Lock for setting :attr:`.last_opencv_init`


        .. note::

            OpenCV must be installed to use this class! A Prebuilt opencv binary is available for the raspberry pi,
            but it doesn't take advantage of some performance-enhancements available to OpenCV. Use the ``install_opencv.sh``
            script in the setup directory to compile OpenCV with these enhancements.

        If your camera isn't working, to print debugging information you can run::

            echo 3 > /sys/class/video4linux/videox/dev_debug

            # check logs
            dmesg



        """
        if not globals()['OPENCV']:
            ImportError('opencv was not imported, and is required for Camera_CV')

        super(Camera_CV, self).__init__(**kwargs)

        self._v4l_info = None

        self.last_opencv_init = globals()['OPENCV_LAST_INIT_TIME']
        self.last_init_lock = globals()['LAST_INIT_LOCK']

        self.camera_idx = camera_idx

    @property
    def fps(self):
        """
        Attempts to get FPS with ``cv2.CAP_PROP_FPS``, uses 30fps as a default

        Returns:
            int: framerate
        """
        fps = self.cam.get(cv2.CAP_PROP_FPS)
        if fps == 0:
            fps = 30
            warnings.warn('Couldnt get fps from camera, using {} as default'.format(fps))
        return fps

    @property
    def shape(self):
        """
        Attempts to get image shape from ``cv2.CAP_PROP_FRAME_WIDTH`` and ``HEIGHT``
        Returns:
            tuple: (width, height)
        """
        if self.crop:
            return (self.crop[2], self.crop[3])
        else:
            return (self.cam.get(cv2.CAP_PROP_FRAME_WIDTH),
                    self.cam.get(cv2.CAP_PROP_FRAME_HEIGHT))

    @shape.setter
    def shape(self, shape):
        """
        Do nothing
        """
        pass

    def _grab(self):
        """
        Reads a frame with :meth:`.cam.read`

        Returns:
            tuple: (timestamp, frame)
        """
        ret, frame = self.cam.read()
        if not ret:
            return False, False
        ts = self._timestamp()
        if self.crop:
            frame = frame[self.crop[1]:self.crop[1]+self.crop[3], self.crop[0]:self.crop[0]+self.crop[2]]
        return (ts, frame)

    def _timestamp(self, frame=None):
        """
        Attempts to get timestamp with ``cv2.CAP_PROP_POS_MSEC``.
        Frame does not need to be passed to this method, as
        timestamps are retrieved from :attr:`.cam`

        .. todo::

            Convert this float timestamp to an isoformatted system timestamp

        Returns:
            float: milliseconds since capture start
        """
        return self.cam.get(cv2.CAP_PROP_POS_MSEC)

    @property
    def backend(self):
        """
        capture backend used by OpenCV for this camera

        Returns:
            str: name of capture backend used by OpenCV for this camera
        """
        return self.cam.getBackendName()

    def init_cam(self):
        """
        Initializes OpenCV Camera

        To avoid overlapping resource allocation requests,
        checks the last time any :class:`.Camera_CV` object was instantiated
        and makes sure it has been at least 2 seconds since then.

        Returns:
            :class:`cv2.VideoCapture`: camera object
        """
        self.initialized.set()

        with self.last_init_lock:
            time_since_last_init = time.time() - self.last_opencv_init.value
            if time_since_last_init < 2.:
                time.sleep(2.0 - time_since_last_init)
            vid = cv2.VideoCapture(self.camera_idx)
            self.last_opencv_init.value = time.time()

        self.logger.info("Camera Initialized")

        return vid

    def release(self):
        self.stop()
        self.cam.release()
        self._cam = None
        self.initialized.clear()
        super(Camera_CV, self).release()


    @property
    def v4l_info(self):
        """
        Device information from ``v4l2-ctl``

        Returns:
            dict: Information for all devices available through v4l2
        """
        if not self._v4l_info:
            # query v4l device info
            cmd = ["/usr/bin/v4l2-ctl", '-D']
            out, err = Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()
            out, err = out.strip(), err.strip()

            # split by \n to get lines, then group by \t
            out = out.split('\n')
            out_dict = {}
            vals = {}
            key = ''
            n_indents = 0
            for l in out:

                # if we're a sublist, but not a subsublist, split and strip, make a subdictionary
                if l.startswith('\t') and not l.startswith('\t\t'):
                    this_list = [k.strip() for k in l.strip('\t').split(':')]
                    subkey, subval = this_list[0], this_list[1]
                    vals[subkey] = subval

                # but if we're a subsublist... shouldn't have a dictionary
                elif l.startswith('\t\t'):
                    if not isinstance(vals[subkey], list):
                        # catch the previously assined value from the top level of the subdictionary
                        vals[subkey] = [vals[subkey]]

                    vals[subkey].append(l.strip('\t'))

                else:
                    # otherwise if we're at the bottom level, stash the key and any value dictioanry we've gathered before
                    key = l.strip(':')
                    if vals:
                        out_dict[key] = vals
                        vals = {}

            # get the last one
            out_dict[key] = vals

            self._v4l_info = out_dict

        return self._v4l_info


class Camera_Spinnaker(Camera):

    type="CAMERA_SPIN"

    # only create class attributes if pyspin is detected,
    # otherwise can't import this module without having pyspin


    ATTR_TYPES = {} #: Conversion from data types to pointer types
    ATTR_TYPE_NAMES = {} #: Conversion from data types to human-readable names
    RW_MODES = {} #: Conversion from read/write mode to {'read':bool, 'write':bool} descriptor

    if PYSPIN:
        ATTR_TYPES = {
            PySpin.intfIFloat      : PySpin.CFloatPtr,
            PySpin.intfIBoolean    : PySpin.CBooleanPtr,
            PySpin.intfIInteger    : PySpin.CIntegerPtr,
            PySpin.intfIEnumeration: PySpin.CEnumerationPtr,
            PySpin.intfIString     : PySpin.CStringPtr,
        }

        ATTR_TYPE_NAMES = {
            PySpin.intfIFloat      : 'float',
            PySpin.intfIBoolean    : 'bool',
            PySpin.intfIInteger    : 'int',
            PySpin.intfIEnumeration: 'enum',
            PySpin.intfIString     : 'string',
            PySpin.intfICommand    : 'command',
        }

        RW_MODES = {
            PySpin.RO: {'read':True, 'write': False},
            PySpin.RW: {'read': True, 'write': False},
            PySpin.WO: {'read': False, 'write': False},
            PySpin.NA: {'read': False, 'write': False}
        }


    def __init__(self, serial=None, camera_idx=None, **kwargs):
        """
        Capture video from a FLIR brand camera with the Spinnaker SDK.

        Args:
            serial (str): Serial number of desired camera
            camera_idx (int): If no serial provided, select camera by index. Using ``serial`` is HIGHLY RECOMMENDED.
            **kwargs: passed to :class:`.Camera` metaclass

        .. note::

            PySpin and the Spinnaker SDK must be installed to use this class. Please use the
            ``install_pyspin.sh`` script in ``setup``

        See the documentation for the Spinnaker SDK and PySpin here:

        `<https://www.flir.com/products/spinnaker-sdk/>`_



        Attributes:
            serial (str): Serial number of desired camera
            camera_idx (int): If no serial provided, select camera by index. Using ``serial`` is HIGHLY RECOMMENDED.
            system (:class:`PySpin.System`): The PySpin System object
            cam_list (:class:`PySpin.CameraList`): The list of PySpin Cameras available to the system
            nmap: A reference to the nodemap from the GenICam XML description of the device
            base_path (str): The directory and base filename that images will be written to if object is :attr:`.writing`. eg::

                base_path = '/home/user/capture_directory/capture_'
                image_path = base_path + 'image1.png'

            img_opts (:class:`PySpin.PNGOption`): Options for saving .png images, made by :meth:`~Camera_Spinnaker.write`
        """

        if not PYSPIN:
            raise ImportError('PySpin was not imported, and is required for Camera_Spinnaker')



        self.system = None #spinnaker system
        self.cam_list = None
        self.nmap = None

        self.base_path = None
        self.img_opts = None

        # internal variables
        self._bin = None
        self._exposure = None
        self._fps = None
        self._frame_trigger = None
        self._pixel_format = None
        self._acquisition_mode = None
        self._camera_attributes = {}
        self._camera_methods = {}
        self._camera_node_types = {}
        self._readable_attributes = {}
        self._writable_attributes = {}
        self._timestamps = []


        super(Camera_Spinnaker, self).__init__(**kwargs)

        if serial and camera_idx:
            self.logger.warning("serial and camera_idx were both passed, defaulting to serial")
            camera_idx = None

        if isinstance(serial, float) or isinstance(serial, int):
            serial = str(serial)
        self.serial = serial
        self.camera_idx = camera_idx



        # set passed parameters
        # has to be done in a specific order, as they are mutually dependent.
        # eg. exposure depends on fps, which depends on bin, etc.
        if 'pixel_format' in kwargs.keys():
            self.set('PixelFormat', kwargs['pixel_format'])
        else:
            try:
                self.set('PixelFormat', PySpin.PixelFormat_Mono8)
            except:
                pass

        if 'bin' in kwargs.keys():
            self.bin = kwargs['bin']
        if 'fps' in kwargs.keys():
            self.fps = kwargs['fps']

        if 'acquisition_mode' in kwargs.keys():
            self.set('AcquisitionMode', kwargs['acquisition_mode'])
        else:
            self.acquisition_mode = 'continuous'



    def init_cam(self):
        """
        Initialize the Spinnaker Camera

        Initializes the camera, system, cam_list, node map, and the camera methods and
        attributes used by :meth:`~Camera_Spinnaker.get` and :meth:`~Camera_Spinnaker.set`

        Returns:
            :class:`PySpin.Camera`: The Spinnaker camera object
        """

        # find our camera!
        # get the spinnaker system handle
        self.system = PySpin.System.GetInstance()
        # need to hang on to camera list for some reason, could be cargo cult code
        self.cam_list = self.system.GetCameras()


        if self.serial:
            cam = self.cam_list.GetBySerial(self.serial)
        elif self.camera_idx:
            self.logger.warning(
                'No camera serial number provided. \nAddressing cameras by serial number is STRONGLY recommended to avoid randomly using the wrong one')
            self.serial = 'noserial'
            cam = self.cam_list.GetByIndex(self.camera_idx)
        else:
            self.logger.warning(
                'No camera serial number OR camera index provided. Trying to use the first camera. This is a really bad way to call this object'
            )
            cam = self.cam_list.GetByIndex(0)

        # initialize the cam - need to do this before messing w the values
        cam.Init()
        # TODO: Document what a nodemap is...
        self.nmap = cam.GetTLDeviceNodeMap()

        # get list of camera methods and attributes for use with 'get' and 'set' methods
        for node in cam.GetNodeMap().GetNodes():
            pit = node.GetPrincipalInterfaceType()
            name = node.GetName()
            self._camera_node_types[name] = self.ATTR_TYPE_NAMES.get(pit, pit)
            if pit == PySpin.intfICommand:
                self._camera_methods[name] = PySpin.CCommandPtr(node)
            if pit in self.ATTR_TYPES:
                self._camera_attributes[name] = self.ATTR_TYPES[pit](node)

        return cam

    def capture_init(self):
        """
        Prepare the camera for acquisition

        calls the camera's ``BeginAcquisition`` method and populate :attr:`.shape`
        """


        self.cam.BeginAcquisition()
        self.frame = self._grab()
        # FIXME: I think this will break single-shot or multishot modes.
        self.shape = self.frame[1].GetNDArray().shape


    def capture_deinit(self):
        """
        De-initializes the camera after acquisition
        """
        self.cam.EndAcquisition()

    def _process(self):
        """
        Modification of the :meth:`.Camera._process` method for Spinnaker cameras

        Because the objects returned from the :meth:`~Camera_Spinnaker._grab` method are image *pointers*
        rather than :class:`numpy.ndarray`s, they need to be handled differently.

        More details on the differences are given in the :meth:`_write_frame`,
        """
        frame_array = None
        try:
            self.frame = self._grab()
        except Exception as e:
            self.logger.exception(e)

        #self._frame[:] = self.frame[1].GetNDArray()

        if self.writing.is_set():
            self._write_frame()

        if self.streaming.is_set():
            if not frame_array:
                frame_array = self.frame[1].GetNDArray()
            self._stream_q.put_nowait({'timestamp': self.frame[0],
                                       self.name  : frame_array})

        if self.queueing.is_set():
            if not frame_array:
                frame_array = self.frame[1].GetNDArray()
            self.q.put_nowait((self.frame[0], frame_array))

        if self.indicating.is_set():
            if self._indicator is None:
                self._indicator = tqdm()
            self._indicator.update()


        self.frame[1].Release()

    def _grab(self):
        """
        Get next timestamp and PySpin Image

        Returns:
            tuple: (timestamp, :class:`PySpin.Image`)
        """
        img = self.cam.GetNextImage()
        return (self._timestamp(img), img)

    def _timestamp(self, frame=None):
        """
        Get the timestamp from the passed image

        Args:
            frame (:class:`PySpin.Image`): Currently grabbed image

        Returns:
            float: PySpin timestamp
        """
        return frame.GetTimeStamp()


    def write(self, output_filename = None, timestamps=True, blosc=True):
        """
        Sets camera to save acquired images to a directory for later encoding.

        For performance, rather than encoding during acquisition, save each image as
        a (lossless) .png image in a directory generated by :attr:`.output_filename`.

        After capturing is complete, a :class:`.Directory_Writer` encodes the images to an
        x264 encoded .mp4 video.

        Args:
            output_filename (str): Directory to write images to. If None (default), generated by :attr:`.output_filename`
            timestamps (bool): Not used, timestamps are always appended to filenames.
            blosc (bool): Not used, images are directly saved.
        """
        if not output_filename:
            output_filename = self.output_filename
        else:
            self.output_filename = output_filename


        # PNG images are losslessly compressed
        self.img_opts = PySpin.PNGOption()
        self.img_opts.compressionLevel = 1

        # make directory
        output_dir = os.path.splitext(self.output_filename)[0]
        os.makedirs(output_dir)

        # create base_path for output images
        self.base_path = os.path.join(output_dir, "capture_{}__".format(self.name))

        self.writing.set()


    def _write_frame(self):
        """
        Write frame to :attr:`.base_path` + timestamp + '.png' with :meth:`PySpin.Image.Save`
        """
        self.frame[1].Save(self.base_path+str(self.frame[0])+'.png', self.img_opts)


    def _write_deinit(self):
        """
        After capture, write images in :attr:`.base_path` to video with :class:`.Directory_Writer`

        Camera object will remain open until writer has finished.
        """
        self.logger.info('Writing images in {} to {}'.format(self.base_path, self.base_path + '.mp4'))
        self.writer = Directory_Writer(self.base_path, fps=self.fps)
        self.writer.encode()

    @property
    def bin(self):
        """
        Camera Binning.

        Attempts to bin on-device, and use averaging if possible. If averaging not available,
        uses summation.

        Args:
            tuple: tuple of integers, (Horizontal, Vertical binning)

        Returns:
            tuple: (Horizontal, Vertical binning)
        """
        return (int(self.cam.BinningHorizontal.ToString()),
                int(self.cam.BinningVertical.ToString()))

    @bin.setter
    def bin(self, bin):
        self.cam.BinningSelector.SetValue(PySpin.BinningSelector_All)
        try:
            self.cam.BinningHorizontalMode.SetValue(PySpin.BinningHorizontalMode_Average)
            self.cam.BinningVerticalMode.SetValue(PySpin.BinningVerticalMode_Average)
        except PySpin.SpinnakerException:
            self.logger.warning('Average binning not supported, using sum')

        self.cam.BinningHorizontal.SetValue(int(bin[0]))
        self.cam.BinningVertical.SetValue(int(bin[1]))


    @property
    def exposure(self):
        """
        Set Exposure of camera

        Can be set with

        * ``'auto'`` - automatic exposure control. note that this will limit framerate
        * ``float`` from 0-1 - exposure duration proportional to fps. eg. if fps = 10, setting exposure = 0.5 means exposure will be set as 50ms
        * ``float`` or ``int``  >1 - absolute exposure time in microseconds

        Returns:
            str, float: If exposure has been set, return set value. Otherwise return ``.get('ExposureTime')``
        """
        if not self._exposure:
            self._exposure = self.get('ExposureTime')
        return self._exposure

    @exposure.setter
    def exposure(self, exposure):

        if exposure == 'auto':
            self.cam.ExposureAuto.SetValue(PySpin.ExposureAuto_Continuous)
            self._exposure = 'auto'
        elif isinstance(exposure, int) or isinstance(exposure, float):
            if exposure < 1:
                # proportional to fps
                exposure = (1.0 / self.fps) * exposure * 1e6

            try:
                self.set('GainAuto', 'Off')
                self.set('Gain', 1)
            except Exception as e:
                self.logger.exception(e)


            self.cam.ExposureAuto.SetValue(PySpin.ExposureAuto_Off)
            # self.cam.ExposureMode.SetValue(PySpin.ExposureMode_Timed)
            self.cam.ExposureTime.SetValue(exposure)
            self._exposure = exposure
        else:
            self.logger.exception('Dont know how to set exposure {}'.format(exposure))


    @property
    def fps(self):
        """
        Acquisition Framerate

        Set with integer. If set with None, ignored (superclass sets FPS to None on init)

        Returns:
            int: from ``cam.AcquisitionFrameRate.GetValue()``
        """
        return self.cam.AcquisitionFrameRate.GetValue()

    @fps.setter
    def fps(self, fps):
        if isinstance(fps, int):
            #self.cam.AcquisitionFrameRateEnable.SetValue(True)
            #self.cam.AcquisitionFrameRate.SetValue(fps)
            self.set('AcquisitionFrameRateAuto', 'Off')
            self.set('AcquisitionFrameRateEnable',True)
            self.set('AcquisitionFrameRate',fps)

        elif fps is None:
            # initially set to None on superclass init
            pass
        else:
            self.logger.exception('Need to set FPS with an integer')

    @property
    def frame_trigger(self):
        """
        Set camera to lead or follow hardware triggers

        If ``'lead'``, Camera will send TTL pulses from Line 2.

        If ``'follow'``, Camera will follow triggers from Line 3.

        .. seealso::

            * `<https://www.flir.com/support-center/iis/machine-vision/application-note/configuring-synchronized-capture-with-multiple-cameras>`_
            * `<https://www.flir.com/support-center/iis/machine-vision/knowledge-base/what-external-iidc-trigger-modes-are-supported-by-my-camera/>`_
        """

        return self._frame_trigger

    @frame_trigger.setter
    def frame_trigger(self, frame_trigger):

        # if we're generating the triggers...
        if frame_trigger == "lead":
            self.cam.LineSelector.SetValue(PySpin.LineSelector_Line2)
            self.cam.V3_3Enable.SetValue(True)
        elif frame_trigger == "follow":
            self.cam.TriggerMode.SetValue(PySpin.TriggerMode_Off)
            self.cam.TriggerSource.SetValue(PySpin.TriggerSource_Line3)
            # this article says that setting triggeroverlap is necessary, but not sure what it does
            # http://justinblaber.org/acquiring-stereo-images-with-spinnaker-api-hardware-trigger/
            self.cam.TriggerOverlap.SetValue(PySpin.TriggerOverlap_ReadOut)
            # In continuous mode, each trigger captures one frame8
            self.cam.TriggerSelector.SetValue(PySpin.TriggerSelector_FrameStart)

            self.cam.TriggerMode.SetValue(PySpin.TriggerMode_On)

        self._frame_trigger = frame_trigger

    @property
    def acquisition_mode(self):
        """
        Image acquisition mode

        One of

        * ``'continuous'`` - continuously acquire frame camera
        * ``'single'`` - acquire a single frame
        * ``'multi'`` - acquire a finite number of frames.

        .. warning::

            Only ``'continuous'`` has been tested.
        """
        return self.cam.AcquisitionMode.ToString()

    @acquisition_mode.setter
    def acquisition_mode(self, acquisition_mode):
        if acquisition_mode == 'continuous':
            self.cam.AcquisitionMode.SetValue(PySpin.AcquisitionMode_Continuous)
        elif acquisition_mode.startwsith("single"):
            self.cam.AcquisitionMode.SetValue(PySpin.AcquisitionMode_SingleFrame)
        elif acquisition_mode.startswith("multi"):
            self.cam.AcquisitionMode.SetValue(PySpin.AcquisitionMode_SingleFrame)
        else:
            self.logger.exception('Acquisition mode must be continuous, single, or multi')


    @property
    def readable_attributes(self):
        """
        All device attributes that are currently readable with :meth:`~Camera_Spinnaker.get`

        Returns:
            dict: A dictionary of attributes that are readable and their current values
        """
        if not self._readable_attributes:

            for k, v in self._camera_attributes.items():
                if '_' in k:
                    continue
                if PySpin.IsReadable(v):
                    self._readable_attributes[k] = self.get(k)

        return self._readable_attributes

    @property
    def writable_attributes(self):
        """
        All device attributes that are currently writeable wth :meth:`~Camera_Spinnaker.set`

        Returns:
            dict: A dictionary of attributes that are writeable and their current values
        """
        if not self._writable_attributes:
            for k, v in self._camera_attributes.items():
                if '_' in k:
                    continue
                if PySpin.IsWritable(v):
                    self._writable_attributes[k] = self.get(k)

        return self._writable_attributes


    def get(self, attr):
        """
        Get a camera attribute.

        Any value in :attr:`.readable_attributes` can be read. Attempts to get numeric values
        with ``.GetValue``, otherwise gets a string with ``.ToString``, so be cautious with types.

        If ``attr`` is a method (ie. in ``._camera_methods``, execute the method and return the value

        Args:
            attr (str): Name of a readable attribute or executable method

        Returns:
            float, int, str: Value of ``attr``
        """
        if attr in self._camera_attributes:

            prop = self._camera_attributes[attr]
            if not PySpin.IsReadable(prop):
                self.logger.exception("Camera property '%s' is not readable" % attr)

            if hasattr(prop, "GetValue"):
                return prop.GetValue()
            elif hasattr(prop, "ToString"):
                return prop.ToString()
            else:
                self.logger.exception("Camera property '%s' is not readable" % attr)
        elif attr in self._camera_methods:
            return self._camera_methods[attr].Execute
        else:
            raise AttributeError(attr)

    def set(self, attr, val):
        """
        Set a camera attribute

        Any value in :attr:`.writeable_attributes` can be set. If attribute has a ``.SetValue`` method,
        (ie. accepts numeric values), attempt to use it, otherwise use ``.FromString``.

        Args:
            attr (str): Name of attribute to be set
            val (str, int, float): Value to set attribute
        """
        if self.cam:
            # checking ensures we have camera initialized
            if attr in self._camera_attributes:

                prop = self._camera_attributes[attr]
                if not PySpin.IsWritable(prop):
                    self.logger.exception("Property '%s' is not currently writable!" % attr)

                if hasattr(prop, 'SetValue'):
                    prop.SetValue(val)
                else:
                    prop.FromString(val)

            elif attr in self._camera_methods:
                self.logger.exception("Camera method '%s' is a function -- you can't assign it a value!" % attr)
            else:
                self.logger.exception('Not sure what to do with attr: {}, value: {}'.format(attr, val))

        # else:
        #
        #     self.__setattr__(attr, val)


    def list_options(self, name):
        """
        List the possible values of a camera attribute.

        Args:
            name (str): name of attribute to query

        Returns:
            dict: Dictionary with {available options: descriptions}
        """
        entries = {}

        if name in self._camera_attributes:
            node = self._camera_attributes[name]
        elif name in self._camera_methods:
            node = self._camera_methods[name]
        else:
            raise ValueError("'%s' is not a camera method or attribute" % name)

        access = False
        if hasattr(node, 'GetAccessMode'):
            access = node.GetAccessMode()

        # print(info)
        if access:
            if hasattr(node, 'GetEntries'):

                for entry in node.GetEntries():
                    entries[entry.GetName().lstrip('EnumEntry_')] = entry.GetDescription().strip()

        else:
            self.logger.exception("Couldn't access attribute {}".format(name))

        return entries

    @property
    def device_info(self):
        """
        Get all information about the camera

        Note that this is distinct from camera *attributes* like fps, instead
        this is information like serial number, version, firmware revision, etc.

        Returns:
            dict: {feature name: feature value}
        """

        device_info = PySpin.CCategoryPtr(self.nmap.GetNode('DeviceInformation'))

        features = device_info.GetFeatures()

        # save information to a dictionary
        info_dict = {}
        for feature in features:
            node_feature = PySpin.CValuePtr(feature)
            info_dict[node_feature.GetName()] = node_feature.ToString()

        return info_dict

    def release(self):
        """
        Release all PySpin objects and wait on writer, if still active.
        """

        super(Camera_Spinnaker, self).release()

        self.stopping.set()

        self._camera_attributes = {}
        self._camera_methods = {}
        self._camera_node_types = {}

        if self.node:
            self.node.release()

        try:
            self.cam.EndAcquisition()
        except Exception as e:
            print(e)

        try:
            self.cam.DeInit()
        except AttributeError:
            pass
        except Exception as e:
            print(e)
            traceback.print_exc(file=sys.stdout)

        try:
            del self._cam
        except AttributeError as e:
            pass
        except Exception as e:
            self.logger.exception(e)

        try:
            self.cam_list.Clear()
            del self.cam_list
        except AttributeError:
            pass
        except Exception as e:
            self.logger.exception(e)

        try:
            del self.nmap
        except AttributeError:
            pass
        except Exception as e:
            self.logger.exception(e)

        try:
            self.system.ReleaseInstance()
        except Exception as e:
            self.logger.exception(e)

        if self.writing.is_set():
            self.writer.wait()




#
# class Camera_Picam(Camera):
#     """
#     also can be used w/ picapture
#     https://lintestsystems.com/wp-content/uploads/2016/09/PiCapture-SD1-Documentation.pdf
#     """
#     pass

#
# class FastWriter(io.FFmpegWriter):
#     def __init__(self, *args, **kwargs):
#         super(FastWriter, self).__init__(*args, **kwargs)
#
#
#     def writeFrame(self, im):
#         """Sends ndarray frames to FFmpeg
#         """
#         vid = vshape(im)
#
#         if not self.warmStarted:
#             T, M, N, C = vid.shape
#             self._warmStart(M, N, C, im.dtype)
#
#         #vid = vid.clip(0, (1 << (self.dtype.itemsize << 3)) - 1).astype(self.dtype)
#
#         try:
#             self._proc.stdin.write(vid.tostring())
#         except IOError as e:
#             # Show the command and stderr from pipe
#             msg = '{0:}\n\nFFMPEG COMMAND:\n{1:}\n\nFFMPEG STDERR ' \
#                   'OUTPUT:\n'.format(e, self._cmd)
#             raise IOError(msg)


class Directory_Writer(object):
    IMG_EXTS = ('.png', '.jpg')
    def __init__(self, dir, fps, ext='.png', ffmpeg_bin='ffmpeg'):
        """
        Encode a directory of images to video with ffmpeg

        Images should be named such that they are machine-orderable, eg.::

            # this
            img-001.png, img-002.png, ... img-010.png

            # not this
            img-1.png, img-2.png, ... img-10.png

        Encoding settings are:

        * ``pix_fmt``: ``yuv420p``
        * ``vcodec`` : ``libx264``
        * ``preset``: ``veryfast``

        .. note::

            ffmpeg must be installed to use this object

        Args:
            dir (str): directory of images to encode
            fps (int): framerate of output video
            ext (str): extension of input images
            ffmpeg_bin (str): ffmpeg binary to use, default is to use ffmpeg in ``$PATH``,
                otherwise specific binary can be specified.
        """
        self.dir = dir
        self.fps = fps
        self.ext = ext
        self.ffmpeg_bin = ffmpeg_bin

        self.encode_thread = None

    def encode(self):
        """
        Begin encoding.

        calls :meth:`._encode` in a thread.

        Encoding calls ffmpeg with a glob string like::

            self.dir + '*' + self.ext
        """
        self.encode_thread = threading.Thread(target=self._encode)
        self.encode_thread.start()

    def _encode(self):

        glob_str = os.path.join(self.dir.rstrip(os.sep)+ '*'+self.ext)

        ffmpeg_cmd = [self.ffmpeg_bin, "-y", '-r', str(self.fps),
                      '-pattern_type', 'glob', '-i', glob_str,
                      '-pix_fmt', 'yuv420p', '-r', str(self.fps),
                      '-vcodec', 'libx264', '-preset', 'veryfast',
                      self.dir.rstrip(os.sep).rstrip('__')+'.mp4']

        result = subprocess.call(ffmpeg_cmd)
        return result

    def wait(self):
        """
        ``.join`` the encoding thread.
        """
        if self.encode_thread:
            self.encode_thread.join()



class Video_Writer(mp.Process):
    def __init__(self, q, path, fps=None, timestamps=True, blosc=True):
        """
        Encode frames as they are acquired in a separate process.

        Must call :meth:`~Video_Writer.start` after initialization to begin encoding.

        Encoding continues until 'END' is put in :attr:`~Video_Writer.q`.

        Timestamps are saved in a .csv file with the same path as the video.

        Args:
            q (:class:`~queue.Queue`): Queue into which frames will be dumped
            path (str): output path of video
            fps (int): framerate of output video
            timestamps (bool): if True (default), input will be of form (timestamp, frame). if False,
                input will just be frames and timestamps will be generated as the frame is encoded (**not recommended**)
            blosc (bool): if True, frames in the :attr:`~Video_Writer.q` will be compresed with blosc. if False, uncompressed

        Attributes:
            timestamps (list): Timestamps for frames, written to .csv on completion of encoding

        """

        super(Video_Writer, self).__init__()

        self.q = q
        self.path = path
        self.fps = fps
        self.given_timestamps = timestamps
        self.timestamps = []
        self.blosc = blosc


        if fps is None:
            warnings.warn('No FPS given, using 30fps by default')
            self.fps = 30

    def run(self):
        """
        Open a :class:`skvideo.io.FFmpegWriter` and begin processing frames from :attr:`~Video_Writer.q`

        Should not be called by itself, overwrites the :meth:`multiprocessing.Process.run` method,
        so should call :meth:`Video_Writer.start`

        Continue encoding until 'END' put in queue.
        """

        self.timestamps = []


        out_vid_fn = self.path
        vid_out = io.FFmpegWriter(out_vid_fn,
        #vid_out = FastWriter(out_vid_fn,
            inputdict={
                '-r': str(self.fps),
        },
            outputdict={
                '-vcodec': 'libx264',
                '-pix_fmt': 'yuv420p',
                '-r': str(self.fps),
                '-preset': 'ultrafast',
            },
            verbosity=1
        )

        try:

            for input in iter(self.q.get, 'END'):
                try:

                    if self.given_timestamps:
                        self.timestamps.append(input[0])
                        if self.blosc:
                            vid_out.writeFrame(blosc.unpack_array(input[1]))
                        else:
                            vid_out.writeFrame(input[1])
                    else:
                        self.timestamps.append(datetime.now().isoformat())
                        if self.blosc:
                            vid_out.writeFrame(blosc.unpack_array(input))
                        else:
                            vid_out.writeFrame(input)

                except Exception as e:
                    print(e)
                    traceback.print_tb()
                    # TODO: Too general
                    break

        finally:

            # save timestamps as .csv
            ts_path = os.path.splitext(self.path)[0] + '.csv'
            with open(ts_path, 'w') as ts_file:
                csv_writer = csv.writer(ts_file)
                for ts in self.timestamps:
                    csv_writer.writerow([ts])

            vid_out.close()


def list_spinnaker_cameras():
    """
    List all available Spinnaker cameras and their ``DeviceInformation``

    Returns:
        list: list of dictionaries of device information for each camera.
    """
    system = PySpin.System.GetInstance()
    cam_list = system.GetCameras()

    cam_info = []
    for cam in cam_list:
        nmap = cam.GetTLDeviceNodeMap()

        device_info = PySpin.CCategoryPtr(nmap.GetNode('DeviceInformation'))

        features = device_info.GetFeatures()

        # save information to a dictionary
        info_dict = {}
        for feature in features:
            node_feature = PySpin.CValuePtr(feature)
            info_dict[node_feature.GetName()] = node_feature.ToString()

        cam_info.append(info_dict)

        del cam
    cam_list.Clear()
    system.ReleaseInstance()

    return cam_info


import os
import sys
import typing
from datetime import datetime

import numpy as np

from autopilot import prefs
from autopilot.transform.transforms import Transform


class Image(Transform):
    """
    Metaclass for transformations of images
    """

    def __init__(self, shape=None, *args, **kwargs):
        super(Image, self).__init__(*args, **kwargs)
        self._shape = shape # type: typing.Tuple[int, int]

    @property
    def format_in(self) -> dict:
        return {
            'type': np.ndarray,
            'shape': self.shape
        }


    @format_in.setter
    def format_in(self, format_in: dict):
        if 'shape' in format_in.keys():
            self.shape = format_in['shape']

    @property
    def format_out(self) -> dict:
        raise NotImplementedError('Every subclass of Transform must define format_out!')

    @format_out.setter
    def format_out(self, format_out: dict):
        raise NotImplementedError('Every subclass of Transform must define format_out!')

    @property
    def shape(self) -> typing.Tuple[int, int]:
        return self._shape

    @shape.setter
    def shape(self, shape):
        if not isinstance(shape, tuple) and not len(shape) == 2:
            raise ValueError('shape must be a tuple of two integers')

        self._shape = (round(shape[0]), round(shape[1]))


class DLC(Image):
    """
    Do pose estimation with ``DeepLabCut-Live``!!!!!

    Specify a ``model_dir`` (relative to ``<BASEDIR>/dlc`` or absolute) or a model name from the DLC model zoo.

    All other args and kwargs are passed on to :class:`dlclive.DLCLive`,
    see its documentation for details: https://github.com/DeepLabCut/DeepLabCut-live

    Attributes:
        model_type (str, 'local' or 'zoo'): whether a directory (local) or a modelzoo name (zoo) was passed
        live (:class:`dlclive.DLCLive`): the DLCLive object
    """

    def __init__(self, model_dir: str = None, model_zoo: str = None, *args, **kwargs):
        """
        Must give either ``model_dir`` or ``model_zoo``

        Args:
            model_dir (str): directory of model, either absolute or relative to ``<BASEDIR>/dlc``. if ``None``, use ``model_zoo``
            model_zoo (str): name of modelzoo model. if ``None``, use ``model_dir``

            *args: passed to DLCLive and superclass
            **kwargs: passed to DLCLive and superclass
        """
        super(DLC, self).__init__(*args, **kwargs)

        self._model = None
        self._model_dir = None
        self._deeplabcut = None
        self.model_type = None
        self.initialized = False

        if not model_dir and not model_zoo:
            raise ValueError('Either model_dir or model_zoo must be given!')
        elif model_dir and model_zoo:
            raise ValueError('Only model_dir OR model_zoo can be given!!')

        try:
            from dlclive import DLCLive
        except ImportError as e:
            print('dlclive not found! dlclive must be installed before using this transform')
            raise e

        self.import_dlc()
        self.deeplabcut = sys.modules['deeplabcut']

        if model_dir:
            # figure out what model to use!
            self.model_type = 'local'
            self.model = model_dir
            self.model_dir = model_dir

        elif model_zoo:
            # check if valid
            self.model_type = "zoo"
            self.model = model_zoo
            # self.model_dir = os.path.join(self.dlc_dir, model_zoo)

        #makes sure model is exported, doesn't do anything if already exported

        self.export_model()

        self.live = DLCLive(self.dlc_paths['export_dir'], *args, **kwargs)
        self.live.init_inference()

    def process(self, input: np.ndarray) -> np.ndarray:
        if not self.live.is_initialized:
            output = self.live.init_inference(input)
        else:
            output = self.live.get_pose(input)

        return output

    @property
    def model(self) -> str:
        return self._model

    @model.setter
    def model(self, model: str):
        if self.model_type == 'zoo':
            if model not in self.list_modelzoo():
                raise ValueError(f'model "{model}" not found in available models: {self.list_modelzoo()}')

            # check if we already have the model downloaded
            models = os.listdir(self.dlc_dir)
            models = [m for m in models if m.startswith(model)]
            if len(models) == 0:
                cfg_path = self.create_modelzoo(model)
                self.model_dir = os.path.dirname(cfg_path)
            elif len(models) == 1:
                self.model_dir = os.path.join(self.dlc_dir,models[0])
            else:
                # more than one, pick the most recent one
                most_recent = datetime(1970,1,1)
                for i, test_model_dir in enumerate(models):
                    pieces = test_model_dir.split('-')
                    this_date = datetime(int(pieces[-3]), int(pieces[-2]), int(pieces[-1]))
                    if this_date > most_recent:
                        model_dir = test_model_dir
                        most_recent = this_date
                self.model_dir = model_dir



        elif self.model_type == 'local':
            pass

        self._model = model

    @property
    def model_dir(self) -> str:
        return self._model_dir

    @model_dir.setter
    def model_dir(self, model_dir):
        if self.model_type == 'zoo':
            pass

            # if not os.path.exists(model_dir):
            #
            #     # self.create_modelzoo()

        elif self.model_type == 'local':

            if not os.path.exists(model_dir):
                # if we were given a path that can be resolved, use it
                # otherwise, check if the model is in our DLC_DIR
                autopilot_model_dir = os.path.join(self.dlc_dir, model_dir)
                if os.path.exists(autopilot_model_dir):
                    model_dir = autopilot_model_dir
                else:
                    # see if a model *starting* with the model name exists
                    autopilot_models = os.listdir(self.dlc_dir)
                    autopilot_models = [a for a in autopilot_models if a.startswith(model_dir)]
                    if len(autopilot_models)>0:
                        model_dir = os.path.join(self.dlc_dir,autopilot_models[0])

        if not os.path.exists(model_dir):
            raise ValueError(f'model_dir given, but model could not be found!\nmodel_dir: {model_dir}')

        self._model_dir = model_dir

    @property
    def dlc_paths(self) -> dict:
        """
        paths used by dlc in manipulating/using models

        * config: <model_dir>/config.yaml
        * train_pose_cfg: <model_dir>/dlc-models/iteration-<n>/<name>/train/pose_cfg.yaml,
        * export_pose_cfg: <model_dir>/exported-models/<name>/pose_cfg.yaml
        * export_dir: <model_dir>/exported-models/<name>

        Returns:
            dict
        """
        config = os.path.join(self.model_dir, 'config.yaml')

        # get pose_cfg in training folder
        train_pose_cfg = None
        # get latest iteration
        dlc_models = os.path.join(self.model_dir, 'dlc-models')
        iteration = sorted([it for it in os.listdir(dlc_models) if it.startswith('iteration')])
        if len(iteration)>0:
            # within an iteration, maybe multiple training sessions
            iteration = os.path.join(dlc_models, iteration[-1])
            training_session = sorted(os.listdir(iteration))
            if len(training_session)>0:
                test_train_pose_cfg = os.path.join(iteration, training_session[-1], 'train', 'pose_cfg.yaml')
                if os.path.exists(test_train_pose_cfg):
                    train_pose_cfg = test_train_pose_cfg

        # get exported pose_cfg.yaml
        exported_pose_cfg = None
        exported_dir = os.path.join(self.model_dir, 'exported-models')
        exported_model_dir = None
        if os.path.exists(exported_dir):
            # FIXME: assuming only one exported model for now
            exported_subdirs = sorted([subd for subd in os.listdir(exported_dir) if \
                                       os.path.isdir(os.path.join(exported_dir,subd))])
            test_exported_pose_cfg = os.path.join(exported_dir,
                                                  exported_subdirs[-1],
                                                  'pose_cfg.yaml')
            if os.path.exists(test_exported_pose_cfg):
                exported_pose_cfg = test_exported_pose_cfg
                exported_model_dir = os.path.join(exported_dir, exported_subdirs[-1])


        return {
            'config':config,
            'train_pose_cfg': train_pose_cfg,
            'export_pose_cfg': exported_pose_cfg,
            'export_dir': exported_model_dir
        }

    @property
    def dlc_dir(self) -> str:
        """
        ``{prefs.BASE_DIR}/dlc``
        Returns:
            str
        """
        if 'DLCDIR' in prefs.prefdict.keys():
            dlc_dir = prefs.DLCDIR
        else:
            dlc_dir = os.path.join(prefs.BASEDIR, 'dlc')
            if not os.path.exists(dlc_dir):
                try:
                    os.mkdir(dlc_dir)

                except OSError as e:
                    raise OSError(f'No DLC dir found and one could not be created!\n{e}')
            prefs.add('DLC_DIR', dlc_dir)

        return dlc_dir

    @classmethod
    def list_modelzoo(cls):
        """
        List available modelzoo model names in local deeplabcut version

        Returns:
            list: names of available modelzoo models
        """
        __deeplabcut = __import__('deeplabcut')
        return __deeplabcut.create_project.modelzoo.Modeloptions

    #
    # @property
    # def deeplabcut(self):
    #     if not self._deeplabcut:
    #         try:
    #             os.environ['DLClight'] = "True"
    #             os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    #             self._deeplabcut = __import__('deeplabcut')
    #         except ImportError as e:
    #             print('deeplabcut not found! deeplabcut must be installed before using this transform!')
    #             raise e
    #     return self._deeplabcut

    def import_dlc(cls):
        if 'deeplabcut' not in sys.modules.keys():
            try:
                os.environ['DLClight'] = "True"
                os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
                import deeplabcut
            except ImportError as e:
                print('deeplabcut not found! deeplabcut must be installed before using this transform!')
                raise e

    def create_modelzoo(self, model):
        self.model_type = "zoo"
        cfg_path, _ = self.deeplabcut.create_pretrained_project(
            model,
            'autopilot',
            [],
            model=model,
            working_directory=self.dlc_dir,
            analyzevideo=False,
            createlabeledvideo=False,
            exportmodel=True)
        return cfg_path

    def load_model(self):
        pass

    def export_model(self):
        # pdb.set_trace()
        if not self.dlc_paths['export_pose_cfg']:
            try:
                self.deeplabcut.export_model(self.dlc_paths['config'])
            except FileExistsError:
                pass

    @property
    def format_in(self) -> dict:
        return {
            'type': np.ndarray
        }

    @property
    def format_out(self) -> dict:
        return {
            'type': np.ndarray
        }


import os
import sys
from autopilot import prefs
from autopilot.core.networking import Net_Node
from autopilot.hardware import Hardware
from autopilot.hardware.cameras import Camera

import threading
import time
import struct
from datetime import datetime
import numpy as np
from itertools import product
from scipy.interpolate import griddata

if sys.version_info >= (3,0):
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty

if prefs.AGENT in ['pilot']:
    import pigpio

try:
    import MLX90640 as mlx_cam
    MLX90640_LIB = True
except ImportError:
    MLX90640_LIB = False



class I2C_9DOF(Hardware):
    """
    A `Sparkfun 9DOF<https://www.sparkfun.com/products/13944>`_ combined accelerometer, magnetometer, and gyroscope.

    This device uses I2C, so must be connected accordingly:

    - VCC: 3.3V (pin 2)
    - Ground: (any ground pin
    - SDA: I2C.1 SDA (pin 3)
    - SCL: I2C.1 SCL (pin 5)

    This class uses code from the `Adafruit Circuitfun <https://github.com/adafruit/Adafruit_CircuitPython_LSM9DS1>`_ library,
    modified to use pigpio
    """

    # Internal constants and register values:
    _ADDRESS_ACCELGYRO = 0x6B
    _ADDRESS_MAG = 0x1E
    _XG_ID = 0b01101000
    _MAG_ID = 0b00111101

    # Linear Acceleration: mg per LSB
    _ACCEL_MG_LSB_2G = 0.061
    _ACCEL_MG_LSB_4G = 0.122
    _ACCEL_MG_LSB_8G = 0.244
    _ACCEL_MG_LSB_16G = 0.732

    # Magnetic Field Strength: gauss range
    _MAG_MGAUSS_4GAUSS = 0.14
    _MAG_MGAUSS_8GAUSS = 0.29
    _MAG_MGAUSS_12GAUSS = 0.43
    _MAG_MGAUSS_16GAUSS = 0.58

    # Angular Rate: dps per LSB
    _GYRO_DPS_DIGIT_245DPS = 0.00875
    _GYRO_DPS_DIGIT_500DPS = 0.01750
    _GYRO_DPS_DIGIT_2000DPS = 0.07000

    # Temperature: LSB per degree celsius
    _TEMP_LSB_DEGREE_CELSIUS = 8  # 1C = 8, 25 = 200, etc.

    # Register mapping for accelerometer/gyroscope component
    _REGISTER_WHO_AM_I_XG = 0x0F
    _REGISTER_CTRL_REG1_G = 0x10
    _REGISTER_CTRL_REG2_G = 0x11
    _REGISTER_CTRL_REG3_G = 0x12
    _REGISTER_TEMP_OUT_L = 0x15
    _REGISTER_TEMP_OUT_H = 0x16
    _REGISTER_STATUS_REG = 0x17
    _REGISTER_OUT_X_L_G = 0x18
    _REGISTER_OUT_X_H_G = 0x19
    _REGISTER_OUT_Y_L_G = 0x1A
    _REGISTER_OUT_Y_H_G = 0x1B
    _REGISTER_OUT_Z_L_G = 0x1C
    _REGISTER_OUT_Z_H_G = 0x1D
    _REGISTER_CTRL_REG4 = 0x1E
    _REGISTER_CTRL_REG5_XL = 0x1F
    _REGISTER_CTRL_REG6_XL = 0x20
    _REGISTER_CTRL_REG7_XL = 0x21
    _REGISTER_CTRL_REG8 = 0x22
    _REGISTER_CTRL_REG9 = 0x23
    _REGISTER_CTRL_REG10 = 0x24
    _REGISTER_OUT_X_L_XL = 0x28
    _REGISTER_OUT_X_H_XL = 0x29
    _REGISTER_OUT_Y_L_XL = 0x2A
    _REGISTER_OUT_Y_H_XL = 0x2B
    _REGISTER_OUT_Z_L_XL = 0x2C
    _REGISTER_OUT_Z_H_XL = 0x2D

    _REGISTER_WHO_AM_I_M = 0x0F
    _REGISTER_CTRL_REG1_M = 0x20
    _REGISTER_CTRL_REG2_M = 0x21
    _REGISTER_CTRL_REG3_M = 0x22
    _REGISTER_CTRL_REG4_M = 0x23
    _REGISTER_CTRL_REG5_M = 0x24
    _REGISTER_STATUS_REG_M = 0x27
    _REGISTER_OUT_X_L_M = 0x28
    _REGISTER_OUT_X_H_M = 0x29
    _REGISTER_OUT_Y_L_M = 0x2A
    _REGISTER_OUT_Y_H_M = 0x2B
    _REGISTER_OUT_Z_L_M = 0x2C
    _REGISTER_OUT_Z_H_M = 0x2D

    _REGISTER_CFG_M = 0x30
    _REGISTER_INT_SRC_M = 0x31

    _MAGTYPE = True
    _XGTYPE = False
    _SENSORS_GRAVITY_STANDARD = 9.80665

    # User facing constants/module globals.
    ACCELRANGE_2G = (0b00 << 3)
    ACCELRANGE_16G = (0b01 << 3)
    ACCELRANGE_4G = (0b10 << 3)
    ACCELRANGE_8G = (0b11 << 3)
    MAGGAIN_4GAUSS = (0b00 << 5)  # +/- 4 gauss
    MAGGAIN_8GAUSS = (0b01 << 5)  # +/- 8 gauss
    MAGGAIN_12GAUSS = (0b10 << 5)  # +/- 12 gauss
    MAGGAIN_16GAUSS = (0b11 << 5)  # +/- 16 gauss
    GYROSCALE_245DPS = (0b00 << 3)  # +/- 245 degrees/s rotation
    GYROSCALE_500DPS = (0b01 << 3)  # +/- 500 degrees/s rotation
    GYROSCALE_2000DPS = (0b11 << 3)  # +/- 2000 degrees/s rotation

    def __init__(self, *args, **kwargs):
        super(I2C_9DOF, self).__init__(*args, **kwargs)

        # Initialize the pigpio connection
        self.pig = pigpio.pi()

        # Open I2C buses
        self.accel = self.pig.i2c_open(1, self._ADDRESS_ACCELGYRO)
        self.magnet = self.pig.i2c_open(1, self._ADDRESS_MAG)

        # soft reset & reboot accel/gyro and magnet
        self.pig.i2c_write_byte_data(self.accel, self._REGISTER_CTRL_REG8, 0x05)
        self.pig.i2c_write_byte_data(self.magnet, self._REGISTER_CTRL_REG2_M, 0x0C)

        ## enable continuous collection
        # gyro
        self.pig.i2c_write_byte_data(self.accel, self._REGISTER_CTRL_REG1_G, 0xC0)
        # accelerometer
        self.pig.i2c_write_byte_data(self.accel, self._REGISTER_CTRL_REG5_XL, 0x38)
        self.pig.i2c_write_byte_data(self.accel, self._REGISTER_CTRL_REG6_XL, 0xC0)
        # magnetometer
        self.pig.i2c_write_byte_data(self.magnet, self._REGISTER_CTRL_REG3_M, 0x00)

        # set default ranges for sensors
        self.accel_range = self.ACCELRANGE_2G

        self._accel_mg_lsb = None
        self._mag_mgauss_lsb = None
        self._gyro_dps_digit = None

    @property
    def accel_range(self):
        """The accelerometer range.  Must be one of:
          - :attr:`I2C_9DOF.ACCELRANGE_2G`
          - :attr:`I2C_9DOF.ACCELRANGE_4G`
          - :attr:`I2C_9DOF.ACCELRANGE_8G`
          - :attr:`I2C_9DOF.ACCELRANGE_16G`
        """
        reg = self.pig.i2c_read_byte_data(self.accel, self._REGISTER_CTRL_REG6_XL)
        return (reg & 0b00011000) & 0xFF

    @accel_range.setter
    def accel_range(self, val):
        assert val in (self.ACCELRANGE_2G, self.ACCELRANGE_4G,
                       self.ACCELRANGE_8G, self.ACCELRANGE_16G)
        reg = self.pig.i2c_read_byte_data(self.accel, self._REGISTER_CTRL_REG6_XL)

        reg = (reg & ~(0b00011000)) & 0xFF
        reg |= val

        self.pig.i2c_write_byte_data(self.accel, self._REGISTER_CTRL_REG6_XL, reg)
        if val == self.ACCELRANGE_2G:
            self._accel_mg_lsb = self._ACCEL_MG_LSB_2G
        elif val == self.ACCELRANGE_4G:
            self._accel_mg_lsb = self._ACCEL_MG_LSB_4G
        elif val == self.ACCELRANGE_8G:
            self._accel_mg_lsb = self._ACCEL_MG_LSB_8G
        elif val == self.ACCELRANGE_16G:
            self._accel_mg_lsb = self._ACCEL_MG_LSB_16G


    @property
    def mag_gain(self):
        """The magnetometer gain.  Must be a value of:
          - :attr:`I2C_9DOF.MAGGAIN_4GAUSS`
          - :attr:`I2C_9DOF.MAGGAIN_8GAUSS`
          - :attr:`I2C_9DOF.MAGGAIN_12GAUSS`
          - :attr:`I2C_9DOF.MAGGAIN_16GAUSS`
        """
        reg = self.pig.i2c_read_byte_data(self.magnet, self._REGISTER_CTRL_REG2_M)
        return (reg & 0b01100000) & 0xFF

    @mag_gain.setter
    def mag_gain(self, val):
        assert val in (self.MAGGAIN_4GAUSS, self.MAGGAIN_8GAUSS, self.MAGGAIN_12GAUSS,
                       self.MAGGAIN_16GAUSS)
        reg = self.pig.i2c_read_byte_data(self.magnet, self._REGISTER_CTRL_REG2_M)
        reg = (reg & ~(0b01100000)) & 0xFF
        reg |= val
        self.pig.i2c_write_byte_data(self.magnet, self._REGISTER_CTRL_REG2_M, reg)
        if val == self.MAGGAIN_4GAUSS:
            self._mag_mgauss_lsb = self._MAG_MGAUSS_4GAUSS
        elif val == self.MAGGAIN_8GAUSS:
            self._mag_mgauss_lsb = self._MAG_MGAUSS_8GAUSS
        elif val == self.MAGGAIN_12GAUSS:
            self._mag_mgauss_lsb = self._MAG_MGAUSS_12GAUSS
        elif val == self.MAGGAIN_16GAUSS:
            self._mag_mgauss_lsb = self._MAG_MGAUSS_16GAUSS


    @property
    def gyro_scale(self):
        """The gyroscope scale.  Must be a value of:
          - :attr:`I2C_9DOF.GYROSCALE_245DPS`
          - :attr:`I2C_9DOF.GYROSCALE_500DPS`
          - :attr:`I2C_9DOF.GYROSCALE_2000DPS`
        """
        reg = self.pig.i2c_read_byte_data(self.accel, self._REGISTER_CTRL_REG1_G)
        return (reg & 0b00011000) & 0xFF

    @gyro_scale.setter
    def gyro_scale(self, val):
        assert val in (self.GYROSCALE_245DPS, self.GYROSCALE_500DPS, self.GYROSCALE_2000DPS)
        reg = self.pig.i2c_read_byte_data(self.accel, self._REGISTER_CTRL_REG1_G)

        reg = (reg & ~(0b00011000)) & 0xFF
        reg |= val

        self.pig.i2c_write_byte_data(self.accel, self._REGISTER_CTRL_REG1_G, reg)
        if val == self.GYROSCALE_245DPS:
            self._gyro_dps_digit = self._GYRO_DPS_DIGIT_245DPS
        elif val == self.GYROSCALE_500DPS:
            self._gyro_dps_digit = self._GYRO_DPS_DIGIT_500DPS
        elif val == self.GYROSCALE_2000DPS:
            self._gyro_dps_digit = self._GYRO_DPS_DIGIT_2000DPS


    @property
    def acceleration(self):
        """
        The calibrated  x, y, z acceleration in m/s^2

        Returns:
            accel (tuple): x, y, z acceleration

        """
        # taking some code from the pigpio examples
        # http://abyz.me.uk/rpi/pigpio/code/i2c_ADXL345_py.zip
        # and adapting with the sparkfun code in main docstring
        (s, b) = self.pig.i2c_read_i2c_block_data(self.accel, 0x80 | self._REGISTER_OUT_X_L_XL, 6)
        if s >= 0:
            raw =  struct.unpack('<3h', buffer(b))
            return map(lambda x: x*self._accel_mg_lsb / 1000.0 * self._SENSORS_GRAVITY_STANDARD, raw)

    @property
    def magnetic(self):
        """
        The magnetometer X, Y, Z axis values as a 3-tuple of gauss values.

        Returns:
            (tuple): x, y, z gauss values

        """
        (s, b) = self.pig.i2c_read_i2c_block_data(self.magnet, 0x80 | self._REGISTER_OUT_X_L_M, 6)

        if s >= 0:
            raw = struct.unpack('<3h', buffer(b))
            return map(lambda x: x * self._mag_mgauss_lsb / 1000.0, raw)

    @property
    def gyro(self):
        """
        The gyroscope X, Y, Z axis values as a 3-tuple of
        degrees/second values.
        """
        (s, b) = self.pig.i2c_read_i2c_block_data(self.accel, 0x80 | self._REGISTER_OUT_X_L_G, 6)

        if s>=0:
            raw = struct.unpack('<3h', buffer(b))
            return map(lambda x: x * self._gyro_dps_digit, raw)

    @property
    def temperature(self):
        """
        Returns:
            float: Temperature in Degrees C
        """
        (s, b) = self.pig.i2c_read_i2c_block_data(self.accel, 0x80 | self._REGISTER_TEMP_OUT_L, 2)
        buf = buffer(b)
        temp = ((buf[1] << 8) | buf[0]) >> 4
        temp = self._twos_comp(temp, 12)
        return 27.5 + temp/16

    def _twos_comp(self, val, bits):
        # Convert an unsigned integer in 2's compliment form of the specified bit
        # length to its signed integer value and return it.
        if val & (1 << (bits - 1)) != 0:
            return val - (1 << bits)
        return val

class MLX90640(Camera):
    """
    A MLX90640 Temperature sensor.

    Args:
        fps (int): Acquisition framerate, must be one of :attr:`MLX90640.ALLOWED_FPS`
        integrate_frames (int): Number of frames to average over
        interpolate (int): Interpolation multiplier -- 3 "increases the resolution" 3x
        **kwargs: passed to :class:`.Camera`

    Attributes:
        shape (tuple): :attr:`~MLX90640.SHAPE_SENSOR
        integrate_frames (int): Number of frames to average over
        interpolate (int): Interpolation multiplier -- 3 "increases the resolution" 3x
        _grab_event (:class:`threading.Event`): capture thread sets every time it gets a frame,
            _grab waits every time, keeps us from returning same frame twice

    This device uses I2C, so must be connected accordingly:

    - VCC: 3.3V (pin 2)
    - Ground: (any ground pin
    - SDA: I2C.1 SDA (pin 3)
    - SCL: I2C.1 SCL (pin 5)

    Uses a modified version of the `MLX90640 Library <https://github.com/sneakers-the-rat/mlx90640-library>`_
    that is capable of outputting 64fps. You must install the library separately, see the
    ``setup_mlx90640.sh`` script.

    Capture works a bit differently from other Cameras -- the :meth:`~MLX90640.capture_init` method spawns a
    :meth:`~MLX90640._threaded_capture` thread, which continually puts frames in the :attr:`~MLX90640._frames` array
    which serves as a ring buffer. The :meth:`~MLX90640._grab` method then awaits the :attr:`~MLX90640._grab_event` to
    be set by the capture thread, and when it is set returns the mean across frames of the ring buffer.

    .. note::
        The setup script modifies the systemwide i2c baudrate to 1MHz, which may interfere with other
        I2C devices. It can be returned to 400kHz (default) by editing ``/config/boot.txt`` to read
        ``dtparam=i2c_arm_baudrate=400000``

    """
    type='MLX90640'

    ALLOWED_FPS = (1, 2, 4, 8, 16, 32, 64) #: FPS must be one of these
    SHAPE_SENSOR = (32,24) #: (H, W) Output shape of this sensor is always the same. May differ from :attr:`MLX90640.shape` if interpolate >1

    def __init__(self, fps=64, integrate_frames = 64, interpolate = 3, **kwargs):
        """


        """
        if not MLX90640_LIB:
            ImportError('the MLX90640 library was not found, please use the setup_mlx90640.sh script or install manually')

        super(MLX90640, self).__init__(fps, **kwargs)

        # frame shape from the sensor is always the same
        self.shape_sensor = (32, 24)
        # but output shape is dependent on interpolation
        self.shape = (self.SHAPE_SENSOR[0]*interpolate, self.SHAPE_SENSOR[1]*interpolate)


        self._frame_idx = 0
        self._frames = None
        self._integrate_frames = None
        self._interpolate = None
        self._cap_thread = None

        # capture thread sets every time it gets a frame,
        # _grab waits every time.
        # keeps us from returning same frame twice
        self._grab_event = threading.Event()

        # interpolation properties
        self._grid_x = None
        self._grid_y = None
        self._points = list(product(range(self.shape_sensor[0]),
                                    range(self.shape_sensor[1])))

        # set attributes
        self.integrate_frames = integrate_frames
        self.interpolate = interpolate




    @property
    def fps(self):
        return self._fps

    @fps.setter
    def fps(self, fps):
        if fps not in self.ALLOWED_FPS:
            ValueError('fps must be one of {}, got {}'.format(self.ALLOWED_FPS, fps))

        self._fps = fps
        # resets cam attribute, next time it's called for the fps will be set.
        self.cam.cleanup()
        self._cam = None

    @property
    def integrate_frames(self):
        return self._integrate_frames

    @integrate_frames.setter
    def integrate_frames(self, integrate_frames):
        self._frames = np.zeros((self.shape_sensor[0], self.shape_sensor[1], integrate_frames))
        self._integrate_frames = integrate_frames

    @property
    def interpolate(self):
        return self._interpolate

    @interpolate.setter
    def interpolate(self, interpolate):
        if interpolate is not None:
            self._grid_y, self._grid_x = np.meshgrid(np.linspace(0, 24, 24 * interpolate),
                                                   np.linspace(0, 32, 32 * interpolate))
        self._interpolate = interpolate


    def init_cam(self):
        """
        Set the camera object to use our :attr:`MLX90640.fps`
        """
        return mlx_cam.setup(self.fps)

    def capture_init(self):
        """
        Spawn a :meth:`~MLX90640._threaded_capture` thread
        """
        self._cap_thread = threading.Thread(target=self._threaded_capture)
        self._cap_thread.setDaemon(True)
        self._cap_thread.start()


    def _threaded_capture(self):
        """
        Continually capture frames into the :attr:`~MLX90640._frames` ring buffer

        Stops when :attr:`~MLX90640.stopping` is set.
        """
        while not self.stopping.is_set():

            # store the frame in the ringbuffer
            # image comes in all wonky and this is a weird combo of instance and module methods...
            # in order:
            # get frame, cast as array
            # reshape using fortran order and transpose
            # rotate 90 degrees to get normal orientation.
            self._frames[:, :, self._frame_idx] = np.rot90(
                np.array(
                    self.cam.get_frame()
                ).reshape(
                    (self.shape_sensor[0], self.shape_sensor[1]),
                    order="F").T
            )
            self._grab_event.set()
            self._frame_idx = (self._frame_idx + 1) % self.integrate_frames

    def _grab(self):
        """
        Await the :attr:`~MLX90640._grab_event` and then average over the frames stored in
        :attr:`~MLX90640._frames`

        Returns:
            (:class:`~numpy.ndarray`) Averaged and interpolated frame
        """
        ret = self._grab_event.wait(1)
        if not ret:
            return None

        frame = np.mean(self._frames, axis=2)
        self._grab_event.clear()

        if self.interpolate is not None:
            frame = self.interpolate_frame(frame)

        return self._timestamp(), frame

    def _timestamp(self, frame=None):
        """
        Just gets Python timestamps for now...

        Returns:
            str: Isoformatted timestamp from datetime
        """
        return datetime.now().isoformat()

    def interpolate_frame(self, frame):
        """
        Interpolate frame according to :attr:`~MLX90640.interpolate` using :meth:`scipy.interpolate.griddata`

        Args:
            frame (:class:`numpy.ndarray`): Frame to interpolate

        Returns:
            (:class:`numpy.ndarray`): Interpolated Frame
        """
        return griddata(self._points,
                        frame.flatten(),
                        (self._grid_x, self._grid_y),
                        method='cubic')

    def release(self):
        """
        Stops the capture thread, cleans up the camera, and calls the superclass release method.
        """
        self.stopping.set()
        self.cam.cleanup()
        self._cam = None
        super(MLX90640, self).release()












import subprocess
import os
import sys
from autopilot import prefs
import atexit
from time import sleep

PIGPIO = False
try:
    from autopilot.external import pigpio
    PIGPIO = True

except ImportError:
    pass

JACKD = False
# JACKD_MODULE = None # Whether jackd is a module in autopilot.external (True) or use system jackd (False)
JACKD_PROCESS = None
try:
    import jack
    JACKD = True
    # from autopilot.external import jack as autopilot_jack
    #
    # # set env variables
    # jackd_path = os.path.join(autopilot_jack.__path__._path[0])
    #
    # # specify location of libraries when starting jackd
    #
    #
    # if 'LD_LIBRARY_PATH' in os.environ.keys():
    #
    #     os.environ['LD_LIBRARY_PATH'] = ":".join([os.path.join(jackd_path, 'lib'),
    #                                               os.environ.get('LD_LIBRARY_PATH',"")])
    # else:
    #     os.environ['LD_LIBRARY_PATH'] = os.path.join(jackd_path, 'lib')git pu
    # # lib_string = "LD_LIBRARY_PATH=" + os.path.join(jackd_path, 'lib')
    #
    # # specify location of drivers when starting jackd
    # os.environ['JACK_DRIVER_DIR'] = os.path.join(jackd_path, 'lib', 'jack')
    # # driver_string = "JACK_DRIVER_DIR=" + os.path.join(jackd_path, 'lib', 'jack')
    #
    # JACKD = True
    # JACKD_MODULE = True

except (ImportError, OSError):
    pass
    # # try to import jack client, it will look for system jack by default
    # try:
    #     import jack
    #     JACKD = True
    #     JACKD_MODULE = False
    # except OSError:
    #     # no need to do anything, just setting module variables that test what the system is configured to do
    #     pass


def start_pigpiod():
    if not PIGPIO:
        raise ImportError('pigpio was not found in autopilot.external')
    launch_pigpiod = os.path.join(pigpio.__path__._path[0], 'pigpiod')
    if hasattr(prefs, 'PIGPIOARGS'):
        launch_pigpiod += ' ' + prefs.PIGPIOARGS

    if hasattr(prefs, 'PIGPIOMASK'):
        # if it's been converted to an integer, convert back to a string and zfill any leading zeros that were lost
        if isinstance(prefs.PIGPIOMASK, int):
            prefs.PIGPIOMASK = str(prefs.PIGPIOMASK).zfill(28)
        launch_pigpiod += ' -x ' + prefs.PIGPIOMASK

    proc = subprocess.Popen('sudo ' + launch_pigpiod, shell=True)

    # kill process when session ends
    atexit.register(lambda pigpio_proc=proc: pigpio_proc.kill())

    # sleep to let it boot up
    sleep(1)

    return proc

def start_jackd():
    if not JACKD:
        raise ImportError('jackd was not found in autopilot.external or as a system install')

    # get specific launch string from prefs
    if hasattr(prefs, "JACKDSTRING"):
        jackd_string = prefs.JACKDSTRING.lstrip('jackd')

    else:
        jackd_string = ""

    # construct rest of launch string!
    # if JACKD_MODULE:
    #     jackd_path = os.path.join(autopilot_jack.__path__._path[0])
    #
    #     # now set as env variables...
    #     # specify location of libraries when starting jackd
    #     # lib_string = "LD_LIBRARY_PATH=" + os.path.join(jackd_path, 'lib')
    #     #
    #     # specify location of drivers when starting jackd
    #     # os.environ['JACK_DRIVER_DIR'] = os.path.join(jackd_path, 'lib', 'jack')
    #     # driver_string = "JACK_DRIVER_DIR=" + os.path.join(jackd_path, 'lib', 'jack')
    #
    #     jackd_bin = os.path.join(jackd_path, 'bin', 'jackd')
    #
    #     # combine all the pieces
    #     # launch_jackd = " ".join([lib_string, driver_string, jackd_bin, jackd_string])
    #
    # else:
    jackd_bin = 'jackd'

        # launch_jackd = " ".join([jackd_bin, jackd_string])

    launch_jackd = " ".join([jackd_bin, jackd_string])

    proc = subprocess.Popen(launch_jackd, shell=True)
    globals()['JACKD_PROCESS'] = proc

    # kill process when session ends
    atexit.register(lambda jackd_proc=proc: jackd_proc.kill())

    # sleep to let it boot
    sleep(1)

    return proc




"""

Classes that manage hardware logic.

Each hardware class should be able to operate independently - ie. not
be dependent on a particular task class, etc. Other than that there are
very few design requirements:

* Every class should have a .release() method that releases any system
  resources in use by the object, eg. objects that use pigpio must have
  their `pigpio.pi` client stopped; LEDs should be explicitly turned off.
* The very minimal class attributes are described in the :class:`Hardware` metaclass.
* Hardware methods are typically called in their own threads, so care should
  be taken to make any long-running operations internally threadsafe.

Note:
    This software was primarily developed for the Raspberry Pi, which
    has `two types of numbering schemes <https://pinout.xyz/#>`_ ,
    "board" numbering based on physical position and "bcm" numbering
    based on the broadcom chip numbering scheme.

    Board numbering is easier to use, but `pigpio <http://abyz.me.uk/rpi/pigpio/>`_
    , which we use as a bridge between Python and the GPIOs, uses the BCM scheme.
    As such each class that uses the GPIOs takes a board number as its argument
    and converts it to a BCM number in the __init__ method.

    If there is sufficient demand to make this more flexible, we can implement
    an additional `pref` to set the numbering scheme, but the current solution
    works without getting too muddy.

Warning:
    In order to use pigpio, the pigpio daemon must be running. See `the docs <http://abyz.me.uk/rpi/pigpio/python.html>`_
    Usually :class:`~.core.pilot.Pilot` s should be started by the bash script or systemd service
    generated by :mod:`.setup.setup_pilot`, which starts pigpiod.
"""


from autopilot import prefs
from autopilot.core.networking import Net_Node
from datetime import datetime
import os
import logging

# FIXME: Hardcoding names of metaclasses, should have some better system of denoting which classes can be instantiated
# directly for setup and prefs management.
META_CLASS_NAMES = ['Hardware', 'Camera', 'GPIO', 'Directory_Writer', 'Video_Writer']


# pigpio only uses BCM numbers, we need to translate them
# See https://www.element14.com/community/servlet/JiveServlet/previewBody/73950-102-11-339300/pi3_gpio.png
BOARD_TO_BCM = {
     3: 2,   5: 3,   7: 4,   8: 14, 10: 15,
    11: 17, 12: 18, 13: 27, 15: 22, 16: 23,
    18: 24, 19: 10, 21: 9,  22: 25, 23: 11,
    24: 8,  26: 7,  29: 5,  31: 6,  32: 12,
    33: 13, 35: 19, 36: 16, 37: 26, 38: 20,
    40: 21
}
"""
dict: Mapping from board (physical) numbering to BCM numbering. 

See `this pinout <https://pinout.xyz/#>`_.

Hardware objects take board numbered pins and convert them to BCM 
numbers for use with `pigpio`.
"""

BCM_TO_BOARD = dict([reversed(i) for i in BOARD_TO_BCM.items()])
"""
dict: The inverse of :const:`BOARD_TO_BCM`.
"""


class Hardware(object):
    """
    Generic class inherited by all hardware. Should not be instantiated
    on its own (but it won't do anything bad so go nuts i guess).

    Primarily for the purpose of defining necessary attributes.

    Also defines `__del__` to call `release()` so objects are always released
    even if not explicitly.

    Attributes:
        is_trigger (bool): Is this object a discrete event input device?
            or, will this device be used to trigger some event? If `True`,
            will be given a callback by :class:`.Task`, and :meth:`.assign_cb`
            must be redefined.
        pin (int): The BCM pin used by this device, or None if no pin is used.
        type (str): What is this device known as in `.prefs`? Not required.
        input (bool): Is this an input device?
        output (bool): Is this an output device?
    """
    # metaclass for hardware objects
    is_trigger = False
    pin = None
    type = "" # what are we known as in prefs?
    input = False
    output = False

    def __init__(self, name=None, **kwargs):
        if name:
            self.name = name
        else:
            try:
                self.name = self.get_name()
            except:
                Warning('wasnt passed name and couldnt find from prefs for object: {}'.format(self.__str__))
                self.name = None

        self.logger = None
        self.log_handler = None
        self.log_formatter = None
        self.listens = {}
        self.node = None
        self.init_logging()

    def release(self):
        """
        Every hardware device needs to redefine `release()`, and must

        * Safely unload any system resources used by the object, and
        * Return the object to a neutral state - eg. LEDs turn off.

        When not redefined, a warning is given.
        """
        Exception('The release method was not overridden by the subclass!')

    def assign_cb(self, trigger_fn):
        """
        Every hardware device that is a :attr:`~Hardware.trigger` must redefine this
        to accept a function (typically :meth:`.Task.handle_trigger`) that
        is called when that trigger is activated.

        When not redefined, a warning is given.
        """
        if self.is_trigger:
            Exception("The assign_cb method was not overridden by the subclass!")

    def get_name(self):
        """
        Usually Hardware is only instantiated with its pin number,
        but we can get its name from prefs
        """

        # TODO: Unify identification of hardware types across prefs and hardware objects
        try:
            our_type = prefs.HARDWARE[self.type]
        except KeyError:
            our_type = prefs.HARDWARE[self.__class__.__name__]

        for name, pin in our_type.items():
            if self.pin == pin:
                return name
            elif isinstance(pin, dict):
                if self.pin == pin['pin']:
                    return name

    def init_logging(self):
        """
        Initialize logging to a timestamped file in `prefs.LOGDIR` .

        The logger name will be `'node.{id}'` .
        """
        #FIXME: Just copying and pasting from net node, should implement logging uniformly across hw objects
        timestr = datetime.now().strftime('%y%m%d_%H%M%S')
        log_file = os.path.join(prefs.LOGDIR, '{}_{}.log'.format(self.name, timestr))

        self.logger = logging.getLogger('hardware.{}'.format(self.name))
        self.log_handler = logging.FileHandler(log_file)
        self.log_formatter = logging.Formatter("%(asctime)s %(levelname)s : %(message)s")
        self.log_handler.setFormatter(self.log_formatter)
        self.logger.addHandler(self.log_handler)
        if hasattr(prefs, 'LOGLEVEL'):
            loglevel = getattr(logging, prefs.LOGLEVEL)
        else:
            loglevel = logging.WARNING
        self.logger.setLevel(loglevel)
        self.logger.info('{} Logging Initiated'.format(self.name))

    def init_networking(self, listens=None, **kwargs):
        """
        Spawn a :class:`.Net_Node` to :attr:`Hardware.node` for streaming or networked command

        Args:
            listens (dict): Dictionary mapping message keys to handling methods
            **kwargs: Passed to :class:`.Net_Node`

        Returns:

        """

        if not listens:
            listens = self.listens

        self.node = Net_Node(
            self.name,
            upstream=prefs.NAME,
            port=prefs.MSGPORT,
            listens=listens,
            instance=False,
            **kwargs
            #upstream_ip=prefs.TERMINALIP,
            #daemon=False
        )



    def __del__(self):
        self.release()


"""
Client that dumps samples directly to the jack client with the :mod:`jack` package.
"""




import multiprocessing as mp
import queue as queue
import numpy as np
from copy import copy
from threading import Thread
from itertools import cycle
from queue import Empty
import sys

# importing configures environment variables necessary for importing jack-client module below
from autopilot import external

try:
    import jack
except OSError as e:
    print('jack library not found! sounds unavailable')
    # print(e)
    # install_jack = input('Try to install jackd? (y/n): ')
    # if install_jack == "y":
    #     from autopilot.setup.setup_autopilot import run_script
    #     run_script('jackd')
    #     import jack
    # else:
    #     raise e


from autopilot import prefs

# allows us to access the audio server and some sound attributes
SERVER = None
"""
:class:`.JackClient`: After initializing, JackClient will register itself with this variable.
"""

FS = None
"""
int: Sampling rate of the active server
"""

BLOCKSIZE = None
"""
int: Blocksize, or the amount of samples processed by jack per each :meth:`.JackClient.process` call.
"""

QUEUE = None
"""
:class:`multiprocessing.Queue`: Queue to be loaded with frames of BLOCKSIZE audio.
"""

PLAY = None
"""
:class:`multiprocessing.Event`: Event used to trigger loading samples from `QUEUE`, ie. playing.
"""

STOP = None
"""
:class:`multiprocessing.Event`: Event that is triggered on the end of buffered audio.

Note:
    NOT an event used to stop audio.
"""

Q_LOCK = None
"""
:class:`multiprocessing.Lock`: Lock that enforces a single writer to the `QUEUE` at a time.
"""

CONTINUOUS = None
"""
:class:`multiprocessing.Event`: Event that (when set) signals the sound server should play some sound continuously rather than remain silent by default (eg. play a background sound).

"""

CONTINUOUS_QUEUE = None
"""
:class:`multiprocessing.Queue`: Queue that 
"""

CONTINUOUS_LOOP = None
"""
:class:`multiprocessing.Event`: Event flag that is set when frames dropped into the CONTINUOUS_QUEUE should be looped (eg. in the case of stationary background noise),
otherwise they are played and then discarded (ie. the sound is continuously generating and submitting samples)
"""

class JackClient(mp.Process):
    """
    Client that dumps frames of audio directly into a running jackd client.

    When first initialized, sets module level variables above.

    Attributes:
        name (str): name of client, default "jack_client"
        q (:class:`~.multiprocessing.Queue`): Queue that stores buffered frames of audio
        q_lock (:class:`~.multiprocessing.Lock`): Lock that manages access to the Queue
        play_evt (:class:`multiprocessing.Event`): Event used to trigger loading samples from `QUEUE`, ie. playing.
        stop_evt (:class:`multiprocessing.Event`): Event that is triggered on the end of buffered audio.
        quit_evt (:class:`multiprocessing.Event`): Event that causes the process to be terminated.

        client (:class:`jack.Client`): Client to interface with jackd
        blocksize (int): The blocksize - ie. samples processed per :meth:`.JackClient.process` call.
        fs (int): Sampling rate of client
        zero_arr (:class:`numpy.ndarray`): cached array of zeroes used to fill jackd pipe when not processing audio.
        continuous_cycle (:class:`itertools.cycle`): cycle of frames used for continuous sounds
    """
    def __init__(self, name='jack_client'):
        """
        Args:
            name:
        """
        super(JackClient, self).__init__()

        # TODO: If global client variable is set, just return that one.

        self.name = name
        #self.pipe = pipe
        self.q = mp.Queue()
        self.q_lock = mp.Lock()

        self.play_evt = mp.Event()
        self.stop_evt = mp.Event()
        self.quit_evt = mp.Event()

        # we make a client that dies now so we can stash the fs and etc.
        self.client = jack.Client(self.name)
        self.blocksize = self.client.blocksize
        self.fs = self.client.samplerate
        self.zero_arr = np.zeros((self.blocksize,1),dtype='float32')

        # a few objects that control continuous/background sound.
        # see descriptions in module variables
        self.continuous = mp.Event()
        self.continuous_q = mp.Queue(maxsize=1024)
        self.continuous_loop = mp.Event()
        self.continuous.clear()
        self.continuous_loop.clear()
        self.continuous_started = False

        # store the frames of the continuous sound and cycle through them if set in continous mode
        self.continuous_cycle = None


        # store a reference to us and our values in the module
        globals()['SERVER'] = self
        globals()['FS'] = copy(self.fs)
        globals()['BLOCKSIZE'] = copy(self.blocksize)
        globals()['QUEUE'] = self.q
        globals()['Q_LOCK'] = self.q_lock
        globals()['PLAY'] = self.play_evt
        globals()['STOP'] = self.stop_evt
        globals()['CONTINUOUS'] = self.continuous
        globals()['CONTINUOUS_QUEUE'] = self.continuous_q
        globals()['CONTINUOUS_LOOP'] = self.continuous_loop

    def boot_server(self):
        """
        Called by :meth:`.JackClient.run` to boot the server upon starting the process.

        Activates the client and connects it to the number of outports
        determined by `prefs.NCHANNELS`

        :class:`jack.Client` s can't be kept alive, so this must be called just before
        processing sample starts.
        """

        self.client = jack.Client(self.name)
        self.blocksize = self.client.blocksize
        self.fs = self.client.samplerate
        self.zero_arr = np.zeros((self.blocksize,1),dtype='float32')

        self.client.set_process_callback(self.process)

        self.client.outports.register('out_0')

        self.client.activate()
        target_ports = self.client.get_ports(is_physical=True, is_input=True, is_audio=True)

        if hasattr(prefs, 'OUTCHANNELS'):
            if isinstance(prefs.OUTCHANNELS, list):
                for outchan in prefs.OUTCHANNELS:

                    self.client.outports[0].connect(target_ports[int(outchan)])
            elif isinstance(prefs.OUTCHANNELS, int):
                self.client.outports[0].connect(target_ports[prefs.OUTCHANNELS])
            elif isinstance(prefs.OUTCHANNELS, str):
                try:
                    self.client.outports[0].connect(target_ports[int(prefs.OUTCHANNELS)])
                except TypeError:
                    Exception('Could not coerce prefs.OUTCHANNELS to an integer or list of ints. Connecting to port 0. got {}'.format(prefs.OUTCHANNELS))
                    self.client.outports[0].connect(target_ports[0])
        else:
            self.client.outports[0].connect(target_ports[0])
            if prefs.NCHANNELS == 2:
                # TODO: Limited, obvs. want to handle arbitrary output arrangements.
                self.client.outports[0].connect(target_ports[1])

    def run(self):
        """
        Start the process, boot the server, start processing frames and wait for the end.
        """

        self.boot_server()

        # we are just holding the process open, so wait to quit
        try:
            self.quit_evt.clear()
            self.quit_evt.wait()
        except KeyboardInterrupt:
            # just want to kill the process, so just continue from here
            pass


    # def close(self):
    #     # TODO: shut down server but also reset module level variables
    #     pass

    def quit(self):
        """
        Set the :attr:`.JackClient.quit_evt`
        """
        self.quit_evt.set()

    def process(self, frames):
        """
        Process a frame of audio.

        If the :attr:`.JackClient.play_evt` is not set, fill port buffers with zeroes.

        Otherwise, pull frames of audio from the :attr:`.JackClient.q` until it's empty.

        When it's empty, set the :attr:`.JackClient.stop_evt` and clear the :attr:`.JackClient.play_evt` .

        Warning:
            Handling multiple outputs is a little screwy right now. v0.2 effectively only supports one channel output.

        Args:
            frames: number of frames (samples) to be processed. unused. passed by jack client
        """
        if not self.play_evt.is_set():
            # if we are in continuous mode...
            if self.continuous.is_set():

                try:
                    data = self.continuous_q.get_nowait()

                except Empty:
                    # TODO: Logging for sound client
                    Warning('Continuous queue was empty!')
                    #self.continuous.clear()
                    data = self.zero_arr

                self.client.outports[0].get_array()[:] = data.T

                # if not self.continuous_started:
                #     # if we are just entering continuous mode, get the continuous sound and prepare to play it
                #     continuous_frames = []
                #     while not self.continuous_q.empty():
                #         try:
                #             continuous_frames.append(self.continuous_q.get_nowait())
                #         except Empty:
                #             break
                #     self.continuous_cycle = cycle(continuous_frames)
                #     self.continuous_started = True
                #
                # # FIXME: Multichannel sound....
                # self.client.outports[0].get_array()[:] = self.continuous_cycle.next().T

            else:
                for channel, port in zip(self.zero_arr.T, self.client.outports):
                    port.get_array()[:] = channel
        else:

            try:
                data = self.q.get_nowait()
            except queue.Empty:
                data = None
                Warning('Queue Empty')
            if data is None:
                # fill with continuous noise
                if self.continuous.is_set():
                    #self.client.outports[0].get_array()[:] = self.continuous_cycle.next().T
                    try:
                        data = self.continuous_q.get_nowait()
                    except Empty:
                        # TODO: Logging for sound client
                        Warning('Continuous queue was empty!')
                        # self.continuous.clear()
                        data = self.zero_arr
                        #self.continuous.clear()

                    self.client.outports[0].get_array()[:] = data.T


                else:
                    for channel, port in zip(self.zero_arr.T, self.client.outports):
                        port.get_array()[:] = channel
                # sound is over
                self.play_evt.clear()
                self.stop_evt.set()
            else:
                if data.shape[0] < self.blocksize:
                    # if sound was not padded, fill remaining with continuous sound or silence
                    n_from_end = self.blocksize - data.shape[0]
                    if self.continuous.is_set():
                        # data = np.concatenate((data, self.continuous_cycle.next()[-n_from_end:]),
                        #                       axis=0)
                        cont_data = self.continuous_q.get_nowait()
                        data = np.concatenate((data, cont_data[-n_from_end:]),
                                              axis=0)
                    else:
                        data = np.pad(data, (0, n_from_end), 'constant')
                #TODO: Fix the multi-output situation so it doesn't get all grumbly.
                # use cycle so if sound is single channel it gets copied to all outports

                self.client.outports[0].get_array()[:] = data.T
                #for channel, port in zip(cycle(data.T), self.client.outports):
                #    port.get_array()[:] = channel







"""
After initial setup, configure autopilot: create an autopilot directory and a prefs.json file

"""

import npyscreen as nps
from collections import OrderedDict as odict
import pprint
import json
import os
import subprocess
import argparse
import sys
import pdb
from copy import copy
import inspect
import pkgutil
import ast
import importlib
import threading
import shlex

from autopilot import hardware

# CLI Options
parser = argparse.ArgumentParser(description="Setup an Autopilot Agent")
parser.add_argument('-f', '--prefs', help="Location of .json prefs file (default: ~/autopilot/prefs.json")
parser.add_argument('-d', '--dir', help="Autopilot directory (default: ~/autopilot)")
parser.add_argument('-s', '--script', help="Run a setup script without entering a full setup routine. for available scripts see -l")
parser.add_argument('-l', '--list_scripts', help="list available setup scripts!", action='store_true')

AGENTS = ('TERMINAL', 'PILOT', 'CHILD')


ENV_PILOT = odict({
    'env_pilot'   : {'type': 'bool',
                     'text': 'install system packages necessary for autopilot? (required if they arent already)'},
    'performance' : {'type': 'bool',
                     'text': 'Do performance enhancements? (recommended, change cpu governor and give more memory to audio)'},
    'change_pw': {'type': 'bool',
                  'text': "If you haven't, you should change the default raspberry pi password or you _will_ get your identity stolen. Change it now?"},
    'set_locale': {'type': 'bool',
                   'text': 'Would you like to set your locale?'},
    'hifiberry' : {'type': 'bool',
                   'text': 'Setup Hifiberry DAC/AMP?'},
    'viz'       : {'type': 'bool',
                   'text': 'Install X11 server and psychopy for visual stimuli?'},
    'bluetooth' : {'type': 'bool',
                   'text': 'Disable Bluetooth? (recommended unless you\'re using it <3'},
    'systemd'   : {'type': 'bool',
                   'text': 'Install Autopilot as a systemd service?\nIf you are running this command in a virtual environment it will be used to launch Autopilot'},
    'jackd'     : {'type': 'bool',
                   'text': 'Install jack audio (required if AUDIOSERVER == jack)'}
})

BASE_PREFS = odict({
    'NAME'       : {'type': 'str', "text": "Agent Name:"},
    'BASEDIR'    : {'type': 'str', "text":"Base Directory:", "default":os.path.join(os.path.expanduser("~"),"autopilot")},
    'PUSHPORT'   : {'type': 'int',"text":"Push Port - Router port used by the Terminal or upstream agent:", "default":"5560"},
    'MSGPORT'    : {'type': 'int', "text":"Message Port - Router port used by this agent to receive messages:", "default":"5565"},
    'TERMINALIP' : {'type': 'str', "text":"Terminal IP:", "default":"192.168.0.100"},
    'LOGLEVEL'   : {'type': 'choice', "text": "Log Level:", "choices":("DEBUG", "INFO", "WARNING", "ERROR"), "default": "WARNING"},
    'CONFIG'     : {'type': 'list', "text": "System Configuration", 'hidden': True}
})

PILOT_PREFS = odict({
    'PIGPIOMASK': {'type': 'str', 'text': 'Binary mask controlling which pins pigpio controls according to their BCM numbering, see the -x parameter of pigpiod',
                   'default': "1111110000111111111111110000"},
    'PIGPIOARGS': {'type': 'str', 'text': 'Arguments to pass to pigpiod on startup',
                   'default': '-t 0 -l'},
    'PULLUPS'   : {'type': 'list', 'text': 'Pins to pull up on system startup? (list of form [1, 2]'},
    'PULLDOWNS'   : {'type': 'list', 'text': 'Pins to pull down on system startup? (list of form [1, 2]'}
})

LINEAGE_PREFS = odict({
    'LINEAGE':     {'type': 'choice', "text": "Are we a parent or a child?", "choices": ("NONE", "PARENT", "CHILD")},
    'CHILDID'    : {'type': 'str', "text":"Child ID:", "depends":("LINEAGE", "PARENT")},
    'PARENTID'   : {'type': 'str', "text":"Parent ID:", "depends":("LINEAGE", "CHILD")},
    'PARENTIP'   : {'type': 'str', "text":"Parent IP:", "depends":("LINEAGE","CHILD")},
    'PARENTPORT' : {'type': 'str', "text":"Parent Port:", "depends":("LINEAGE", "CHILD")},
})

AUDIO_PREFS = odict({
    'AUDIOSERVER': {'type': 'bool', 'text':'Enable jack audio server?'},
    'NCHANNELS': {'type': 'int', 'text': "Number of Audio channels", 'default':1, 'depends': 'AUDIOSERVER'},
    'OUTCHANNELS': {'type': 'list', 'text': 'List of Audio channel indexes to connect to', 'default': '[1]', 'depends': 'AUDIOSERVER'},
    'FS': {'type': 'int', 'text': 'Audio Sampling Rate', 'default': 192000, 'depends': 'AUDIOSERVER'},
    'JACKDSTRING': {'type': 'str', 'text': 'Arguments to pass to jackd, see the jackd manpage',
                    'default': 'jackd -P75 -p16 -t2000 -dalsa -dhw:sndrpihifiberry -P -rfs -n3 -s &', 'depends': 'AUDIOSERVER'},
})

TERMINAL_PREFS = odict({
    'DRAWFPS': {'type': 'int', "text": "FPS to draw videos displayed during acquisition",
                "default": "20"},
    'PILOT_DB': {'type': 'str', 'text': "filename to use for the .json pilot_db that maps pilots to subjects (relative to BASEDIR)",
                 "default": "pilot_db.json"}
})

DIRECTORY_STRUCTURE = {
    'DATADIR': 'data',
    'SOUNDDIR': 'sounds',
    'LOGDIR': 'logs',
    'VIZDIR': 'viz',
    'PROTOCOLDIR': 'protocols',
}

PILOT_ENV_CMDS = {
    'performance':
        ['sudo systemctl disable raspi-config',
         'sudo sed -i \'/^exit 0/i echo "performance" | sudo tee /sys/devices/system/cpu/cpu*/cpufreq/scaling_governor\' /etc/rc.local',
         'sudo sh -c "echo @audio - memlock 256000 >> /etc/security/limits.conf"',
         'sudo sh -c "echo @audio - rtprio 75 >> /etc/security/limits.conf"',
         ],
    'performance_cameras':
        [
            "sudo sh -c 'echo options uvcvideo nodrop=1 timeout=10000 quirks=0x80 > /etc/modprobe.d/uvcvideo.conf'",
            "sudo rmmod uvcvideo",
            "sudo modprobe uvcvideo",
            "sudo sed -i \"/^exit 0/i sudo sh -c 'echo ${usbfs_size} > /sys/module/usbcore/parameters/usbfs_memory_mb'\" /etc/rc.local"
        ],
    'change_pw': ['passwd'],
    'set_locale': ['sudo dpkg-reconfigure locales',
                   'sudo dpkg-reconfigure keyboard-configuration'],
    'hifiberry':
        [
            {'command':'sudo adduser pi i2c', 'optional':True},
            'sudo sed -i \'s/^dtparam=audio=on/#dtparam=audio=on/g\' /boot/config.txt',
            'sudo sed -i \'$s/$/\\ndtoverlay=hifiberry-dacplus\\ndtoverlay=i2s-mmap\\ndtoverlay=i2c-mmap\\ndtparam=i2c1=on\\ndtparam=i2c_arm=on/\' /boot/config.txt',
            'echo -e \'pcm.!default {\\n type hw card 0\\n}\\nctl.!default {\\n type hw card 0\\n}\' | sudo tee /etc/asound.conf'
        ],
    'viz': [],
    'bluetooth':
        [
            'sudo sed - i \'$s/$/\ndtoverlay=pi3-disable-bt/\' / boot / config.txt',
            'sudo systemctl disable hciuart.service',
            'sudo systemctl disable bluealsa.service',
            'sudo systemctl disable bluetooth.service'
        ],
    'jackd':
        ['sudo apt update && sudo apt install -y jackd2'],
    'jackd_source':
        [
            "git clone git://github.com/jackaudio/jack2 --depth 1",
            "cd jack2",
            "./waf configure --alsa=yes --libdir=/usr/lib/arm-linux-gnueabihf/",
            "./waf build -j6",
            "sudo ./waf install",
            "sudo ldconfig",
            "sudo sh -c \"echo @audio - memlock 256000 >> /etc/security/limits.conf\"",             # giving jack more juice
            "sudo sh -c \"echo @audio - rtprio 75 >> /etc/security/limits.conf\"",
            "cd ..",
            "rm -rf ./jack2"
        ],
    'opencv':
        [
            "sudo apt-get install -y build-essential cmake ccache unzip pkg-config libjpeg-dev libpng-dev libtiff-dev libavcodec-dev libavformat-dev libswscale-dev libv4l-dev libxvidcore-dev libx264-dev ffmpeg libgtk-3-dev libcanberra-gtk* libatlas-base-dev gfortran python2-dev python-numpy",
            "git clone https://github.com/opencv/opencv.git",
            "git clone https://github.com/opencv/opencv_contrib",
            "cd opencv",
            "mkdir build",
            "cd build",
            "cmake -D CMAKE_BUILD_TYPE=RELEASE \
                -D CMAKE_INSTALL_PREFIX=/usr/local \
                -D OPENCV_EXTRA_MODULES_PATH=/home/pi/git/opencv_contrib/modules \
                -D BUILD_TESTS=OFF \
                -D BUILD_PERF_TESTS=OFF \
                -D BUILD_DOCS=OFF \
                -D WITH_TBB=ON \
                -D CMAKE_CXX_FLAGS=\"-DTBB_USE_GCC_BUILTINS=1 -D__TBB_64BIT_ATOMICS=0\" \
                -D WITH_OPENMP=ON \
                -D WITH_IPP=OFF \
                -D WITH_OPENCL=ON \
                -D WITH_V4L=ON \
                -D WITH_LIBV4L=ON \
                -D ENABLE_NEON=ON \
                -D ENABLE_VFPV3=ON \
                -D PYTHON3_EXECUTABLE=/usr/bin/python3 \
                -D PYTHON_INCLUDE_DIR=/usr/include/python3.7 \
                -D PYTHON_INCLUDE_DIR2=/usr/include/arm-linux-gnueabihf/python3.7 \
                -D OPENCV_ENABLE_NONFREE=ON \
                -D INSTALL_PYTHON_EXAMPLES=OFF \
                -D WITH_CAROTENE=ON \
                -D CMAKE_SHARED_LINKER_FLAGS='-latomic' \
                -D BUILD_EXAMPLES=OFF ..",
            "sudo sed -i 's/^CONF_SWAPSIZE=100/CONF_SWAPSIZE=2048/g' /etc/dphys-swapfile", # increase size of swapfile so multicore build works
            "sudo /etc/init.d/dphys-swapfile stop",
            "sudo /etc/init.d/dphys-swapfile start",
            "make -j4",
            "sudo make install",
            "sudo ldconfig",
            "sudo sed -i 's/^CONF_SWAPSIZE=2048/CONF_SWAPSIZE=100/g' /etc/dphys-swapfile",
            "sudo /etc/init.d/dphys-swapfile stop",
            "sudo /etc/init.d/dphys-swapfile start"
        ],
    'env_pilot':
        [
            "sudo apt-get update",
            "sudo apt-get install -y build-essential cmake git python3-dev libatlas-base-dev libsamplerate0-dev libsndfile1-dev libreadline-dev libasound-dev i2c-tools libportmidi-dev liblo-dev libhdf5-dev libzmq-dev libffi-dev",
        ]

}
"""
performance: 
    * disable startup script that changes cpu governor,
    * change cpu governor to "performance" on boot
    * increase memlock and realtime priority limits for audio group

hifiberry:
    * turn onboard audio off
    * enable hifiberry stuff in /boot/config.txt
    * edit alsa config so hifiberry is default sound card

viz:

.. todo::

    Need to find a more elegant way to do this, for now see lines 160-200 in the presetup_pilot.sh legacy script

"""



class Autopilot_Form(nps.Form):

    def __init__(self, *args, **kwargs):
        self.input = odict()
        self.depends = {}
        super(Autopilot_Form, self).__init__(*args, **kwargs)

        if 'params' in kwargs.keys():
            self.populate_form(kwargs['params'])

    def populate_dependencies(self, params):
        # first find any options that other options depend on
        for param_name, param in params.items():
            if 'depends' in param.keys():
                if isinstance(param['depends'], str):
                    depends_on = param['depends']
                    depend_value = True
                elif isinstance(param['depends'], tuple):
                    depends_on = param['depends'][0]
                    depend_value = param['depends'][1]

                if depends_on in self.depends.keys():
                    self.depends[depends_on].append((param_name, depend_value))
                else:
                    self.depends[depends_on] = [(param_name, depend_value)]

    def populate_form(self, params):

        # check for existing values in global prefs
        global prefs

        self.populate_dependencies(params)

        # create widgets depending on parameter type
        for param_name, param in params.items():
            if param['type'] == 'bool':
                widget = self.add(nps.CheckBox, name=param['text'])
            elif param['type'] in ('str', 'int', 'list'):
                # try to get default from prefs, otherwise use the hardcoded default if present. otherwise blank
                if param_name in prefs.keys():
                    default = prefs[param_name]
                else:
                    try:
                        default = param['default']
                    except KeyError:
                        default = ''

                widget = self.add(nps.TitleText, name=param['text'], value = str(default))
            elif param['type'] == 'choice':
                if param_name in prefs.keys():
                    try:
                        default_ind = [param['choices'].index(prefs[param_name])]
                    except ValueError:
                        default_ind = [0]
                else:
                    try:
                        default = param['default']
                        default_ind = [param['choices'].index(default)]
                    except KeyError:
                        default_ind = [0]

                widget = self.add(nps.TitleSelectOne,
                                  name = param['text'],
                                  values = param['choices'],
                                  max_height=len(param['choices'])+1,
                                  value = default_ind,
                                  scroll_exit = True)
            else:
                raise Warning("Not sure what to do with param {} with type {}".format(param_name, param['type']))

            # if this widget depends on another, initially make it uneditable
            if 'depends' in param.keys():
                widget.editable = False
                widget.color = 'NO_EDIT'
                widget.labelColor = 'NO_EDIT'


            if param_name in self.depends.keys():

                widget.when_value_edited = lambda pname=param_name: self.update_depends(pname)

            self.input[param_name] = widget

    def update_depends(self, param_name):

        #pdb.set_trace()

        # get value
        param_value = self.input[param_name].value
        if hasattr(self.input[param_name], 'values'):
            param_value = self.input[param_name].values[param_value[0]]

        if param_name in self.depends.keys():
            for dependent in self.depends[param_name]:
                if param_value == dependent[1]:
                    self.input[dependent[0]].editable = True
                    self.input[dependent[0]].color = 'DEFAULT'
                    self.input[dependent[0]].labelColor = 'LABEL'
                else:
                    self.input[dependent[0]].editable = False
                    self.input[dependent[0]].color = 'NO_EDIT'
                    self.input[dependent[0]].labelColor = 'NO_EDIT'


class Hardware_Form(nps.FormWithMenus):
    def __init__(self, *args, **kwargs):
        self.input = odict()
        self.altrely = 4
        super(Hardware_Form, self).__init__(*args, **kwargs)


    def create(self):
        self.add(nps.FixedText, value="Use the ctrl+X menu to add new hardware", editable=False, color="VERYGOOD")

        hardware_objs = self.list_hardware()

        for module, hardware_classes in hardware_objs.items():

            mod_menu = self.add_menu(module)
            for class_name in hardware_classes:
                mod_menu.addItem(text=class_name, onSelect=self.add_hardware, arguments=[module, class_name])

    #
    # def init_hardware(self):
    #     global prefs
    #     if 'HARDWARE' in prefs.keys():
    #         for hw_group, hardware_name in prefs
    #
    # def add_hardware_widget(self, sigs):
    #


    def list_hardware(self):
        # start at the top of the autopilot hardware package and work down
        # get all classes that are defined within the hardware module
        base_hardware = [m[0]for m in inspect.getmembers(hardware, inspect.isclass) if m[1].__module__ == hardware.__name__]

        hardware_path = os.path.dirname(hardware.__file__)

        # get names of modules
        submodules = [mod for _, mod, _ in pkgutil.iter_modules([hardware_path])]
        submod_paths = [os.path.join(hardware_path, mod)+'.py' for mod in submodules]

        # we don't want to have to import all the hardware modules just to get a list of them,
        # and only want to try to import them if the user wants to add one to their system
        # so we have to parse the files to get the names of the classes

        hardware_objs = {}
        for submod_name, submod in zip(submodules, submod_paths):
            with open(submod, 'r') as submod_f:
                submod_ast = ast.parse(submod_f.read())

            submod_classes = [n.name for n in submod_ast.body if isinstance(n, ast.ClassDef) and n.name not in hardware.META_CLASS_NAMES]
            hardware_objs[submod_name] = submod_classes

        return hardware_objs




    def add_hardware(self, module, class_name):
        #self.nextrely = 1
        self.DISPLAY()

        # import the class
        hw_class = getattr(importlib.import_module("autopilot.hardware."+module), class_name)
        # get its parent classes (which includes class itself)
        hw_parents = inspect.getmro(hw_class)
        # get signatures for each
        # go in reverse order so top classes options come first
        sigs = []
        # list to keep track of parameter names to remove duplicates
        param_names = []
        for cls in reversed(hw_parents):
            # get signature
            sig = inspect.signature(cls)
            # get parameter names and defaults
            for param_name, param in sig.parameters.items():
                param_default = param.default
                if param_default == inspect._empty:
                    param_default = None
                if param_name in ('kwargs', 'args'):
                    continue

                if param_name not in param_names:
                    # check if we already have a parameter with this name,
                    # if we don't add it.
                    sigs.append((param_name, param_default))
                    param_names.append(param_name)

        MODULE = module.upper()
        # create title and input widgets for arguments

        #pdb.set_trace()

        self.add(nps.FixedText, value="{}.{}".format(module, class_name), rely=self.altrely, editable=False, color="VERYGOOD")

        self.altrely+=1

        hw_widgets = {}
        hw_widgets['type'] = "{}.{}".format(module, class_name)
        for sig in sigs:
            if sig[1] is None:
                sig = (sig[0], '')


            #if isinstance(sig[1], bool):
            #    hw_widgets.append(self.add(nps.CheckBox, name=sig[0], value=sig[1], rely=self.altrely))
            #else:
            hw_widgets[sig[0]] = self.add(nps.TitleText, name=sig[0], value=str(sig[1]), rely=self.altrely)

            self.altrely+=1
        self.altrely+=1

        if MODULE not in self.input.keys():
            self.input[MODULE] = []

        self.input[MODULE].append(hw_widgets)

    def afterEditing(self):
        self.parentApp.setNextForm(None)








class Agent_Form(nps.Form):
    def create(self):
        # self.input = odict({
        #     'AGENT': self.add(nps.TitleSelectOne, max_height=len(AGENTS)+1, value=[0,],
        #                       name="Select an Agent. If this is a Raspberry Pi you should select either 'PILOT' or 'CHILD', and if this is the computer you will be using to control Autopilot you should select 'TERMINAL'", values = AGENTS, scroll_exit=True)
        # })

        if os.path.exists(os.path.join(os.path.dirname(__file__),'welcome_msg.txt')):
            with open(os.path.join(os.path.dirname(__file__),'welcome_msg.txt'), 'r') as welcome_f:
                welcome = welcome_f.read()
                for line in welcome.split('\n'):self.add(nps.FixedText, value=line, editable=False)

        self.input = odict({
            'AGENT': self.add(nps.TitleSelectOne, max_height=len(AGENTS)+1, value=0,
                              name="Select an Autopilot Agent", values=AGENTS, scroll_exit=True)
        })

    def afterEditing(self):
        # terminal
        global prefs

        if self.input['AGENT'].value[0] == 0:
            prefs['AGENT'] = 'TERMINAL'
            self.parentApp.setNextForm('TERMINAL')
        elif self.input['AGENT'].value[0] == 1:
            prefs['AGENT'] = 'PILOT'
            self.parentApp.setNextForm('ENV_PILOT')
        elif self.input['AGENT'].value[0] == 2:
            prefs['AGENT'] = 'CHILD'
            self.parentApp.setNextForm('CHILD')
        else:
            self.parentApp.setNextForm(None)


class Pilot_Env_Form(Autopilot_Form):
    def create(self):
        self.populate_form(ENV_PILOT)

    def afterEditing(self):
        self.parentApp.setNextForm('CONFIG_PILOT_1')

class Pilot_Config_Form_1(Autopilot_Form):
    def create(self):
        self.add(nps.FixedText, value='Base Prefs', editable=False, color="VERYGOOD")
        self.populate_form(BASE_PREFS)
        self.add(nps.FixedText, value='Pilot Prefs', editable=False, color="VERYGOOD")
        self.populate_form(PILOT_PREFS)

    def afterEditing(self):
        self.parentApp.setNextForm('CONFIG_PILOT_2')

class Pilot_Config_Form_2(Autopilot_Form):
    def create(self):
        self.add(nps.FixedText, value='Lineage Prefs', editable=False, color="VERYGOOD")
        self.populate_form(LINEAGE_PREFS)
        self.add(nps.FixedText, value='Audio Prefs', editable=False, color="VERYGOOD")
        self.populate_form(AUDIO_PREFS)

    def afterEditing(self):
        self.parentApp.setNextForm('HARDWARE')



class Terminal_Form(Autopilot_Form):
    def create(self):
        self.add(nps.FixedText, value='Base Prefs', editable=False, color="VERYGOOD")
        self.populate_form(BASE_PREFS)
        self.add(nps.FixedText, value='Terminal Prefs', editable=False, color="VERYGOOD")
        self.populate_form(TERMINAL_PREFS)

    def afterEditing(self):
        self.parentApp.setNextForm(None)



class Autopilot_Setup(nps.NPSAppManaged):
    def __init__(self, prefs):
        super(Autopilot_Setup, self).__init__()
        self.prefs = prefs

    def onStart(self):
        self.agent = self.addForm('MAIN', Agent_Form, name="Select Agent")
        self.env_pilot = self.addForm('ENV_PILOT', Pilot_Env_Form, name="Configure Pilot Environment")
        self.pilot_1 = self.addForm('CONFIG_PILOT_1', Pilot_Config_Form_1, name="Setup Pilot Agent - 1/2")
        self.pilot_2 = self.addForm('CONFIG_PILOT_2', Pilot_Config_Form_2, name="Setup Pilot Agent - 2/2")
        self.hardware = self.addForm('HARDWARE', Hardware_Form, name="Hardware Configuration")
        self.terminal = self.addForm('TERMINAL', Terminal_Form, name="Terminal Configuration")

def unfold_values(v):
    """
    Unfold nested values from the SetupForm. Called recursively.

    Args:
        v (dict): unfolded values
    """
    if isinstance(v, dict):
        # recurse
        v = {k: unfold_values(v) for k, v in v.items()}
    elif isinstance(v, list):
        v = [unfold_values(v) for v in v]
    elif isinstance(v, str):
        # do nothing since this is what we want yno
        pass
    else:
        #pdb.set_trace()

        if hasattr(v, 'values'):
            # if it's an object that has mutiple values, ie. a choice box, value is inside a list.
            v = v.values[v.value[0]]
        elif hasattr(v, 'value'):
            # if it isn't a list, but is still a widget object, get the value
            v = v.value


        try:
            # convert ints to ints, lists to lists, etc. from strings
            v = ast.literal_eval(v)
        except:
            # fine, just a string that can't be evaluated into another type
            pass
    return v

def call_series(commands, series_name=None):
    """
    Call a series of commands, giving a single return code on completion or failure

    :param commands:
    :return:
    """
    if series_name:
        print('\n\033[1;37;42m Running commands for {}\u001b[0m'.format(series_name))

    # have to just combine them -- can't do multiple calls b/c shell doesn't preserve between them
    combined_calls = ""
    last_command = len(commands)-1
    for i, command in enumerate(commands):
        join_with = " && "

        if isinstance(command, str):
            # just a command, default necessary
            combined_calls += command
        elif isinstance(command, dict):
            combined_calls += command['command']

            if command.get('optional', False):
                join_with = "; "

        if i < last_command:
            combined_calls += join_with


    print('Executing:\n    {}'.format(combined_calls))

    result = subprocess.run(combined_calls, shell=True, executable='/bin/bash')

    status = False
    if result.returncode == 0:
        status = True

    if series_name:
        if status:
            print('\n\033[1;37;42m  {} Successful, you lucky duck\u001b[0m'.format(series_name))
        else:
            print('\n\033[1;37;41m  {} Failed, check the error message & ur crystal ball\u001b[0m'.format(series_name))

    return status


def run_script(script_name):
    if script_name in PILOT_ENV_CMDS.keys():
        call_series(PILOT_ENV_CMDS[script_name], script_name)
    else:
        Exception('No script named {}, must be one of {}'.format(script_name, "\n".join(PILOT_ENV_CMDS.keys())))


def list_scripts():
    print('Available Scripts:')
    for script_name in sorted(PILOT_ENV_CMDS.keys()):
        print(f'{script_name}: {PILOT_ENV_CMDS[script_name]["text"]}\n')


def make_dir(adir):
    """
    Make a directory if it doesn't exist and set its permissions to `0777`

    Args:
        adir (str): Path to the directory
    """
    if not os.path.exists(adir):
        os.makedirs(adir)

    os.chmod(adir, 0o774)

if __name__ == "__main__":
    env = {}
    env_params = {}
    prefs = {}
    error_msgs = []
    config_msgs = []

    args = parser.parse_args()

    if args.list_scripts:
        list_scripts()
        sys.exit()
    elif args.script:
        run_script(args.script)
        sys.exit()

    if args.dir:
        autopilot_dir = args.dir

    else:
        # check for a ~/.autopilot file that should point us to the autopilot directory if it exists
        autopilot_conf_fn = os.path.join(os.path.expanduser('~'), '.autopilot')
        if os.path.exists(autopilot_conf_fn):
            with open(autopilot_conf_fn, 'r') as aconf:
                autopilot_dir = aconf.read()
                # autopilot_dir = autopilof_conf['AUTOPILOTDIR']
        else:
            autopilot_dir = os.path.join(os.path.expanduser('~'), 'autopilot', '')

    make_dir(autopilot_dir)


    # attempt to load .prefs from standard location (~/autopilot/prefs.json)
    if args.prefs:
        prefs_fn = args.prefs
    else:
        prefs_fn = os.path.join(autopilot_dir, 'prefs.json')

    if os.path.exists(prefs_fn):
        with open(prefs_fn, 'r') as prefs_f:
            prefs = json.load(prefs_f)

    ###################################3
    # Run the npyscreen prompt

    setup = Autopilot_Setup(prefs)
    setup.run()

    ####################################
    # Collect values

    agent = {k:unfold_values(v) for k, v in setup.agent.input.items()}
    prefs['AGENT'] = agent['AGENT']
    if agent['AGENT'] in ('PILOT', 'CHILD'):
        pilot = odict({k: unfold_values(v) for k, v in setup.pilot_1.input.items()})
        pilot.update(odict({k: unfold_values(v) for k, v in setup.pilot_2.input.items()}))
        hardware_flat = odict({k: unfold_values(v) for k, v in setup.hardware.input.items()})

        # have to un-nest hardware a bit
        # currently is hardware['CAMERAS'] = [{config 1}, {config 2]
        # want         hardware['CAMERAS']['cam_name_1'] = {config}
        hardware = {}
        for hardware_group, hardware_list in hardware_flat.items():
            hardware[hardware_group] = {}
            for hardware_config in hardware_list:
                hardware[hardware_group][hardware_config['name']] = hardware_config

        # get env commands to run
        env_params = odict({k: unfold_values(v) for k, v in setup.env_pilot.input.items()})
        for env_param, result in env_params.items():
            if env_param in PILOT_ENV_CMDS.keys() and result == 1:
                env[env_param] = PILOT_ENV_CMDS[env_param]

        # merge with any existing prefs
        prefs.update(pilot)
        if 'HARDWARE' not in prefs.keys():
            prefs['HARDWARE'] = hardware
        else:
            prefs['HARDWARE'].update(hardware)
    elif agent['AGENT'] == 'TERMINAL':
        # unpack prefs

        terminal = odict({k: unfold_values(v) for k, v in setup.terminal.input.items()})


        # create the pilot_db if it doesn't exist
        terminal['PILOT_DB'] = os.path.join(terminal['BASEDIR'], terminal['PILOT_DB'])
        if not os.path.exists(terminal['PILOT_DB']):
            with open(terminal['PILOT_DB'], 'w') as pilot_db_file:
                json.dump({}, pilot_db_file)

        os.chmod(terminal['PILOT_DB'], 0o775)

        # merge with any existing prefs
        prefs.update(terminal)




    ####################################
    # Configure Environment

    # detect if we are in a virtual environment
    venv_path = ''
    if hasattr(sys, 'real_prefix') or (sys.base_prefix != sys.prefix):
        # virtualenv and pyenv populate these system attrs
        venv_path = sys.prefix
        prefs['VENV'] = venv_path
    else:
        prefs['VENV'] = False

    # get repo directory
    file_loc = os.path.realpath(__file__)
    file_loc = file_loc.split(os.sep)[:-3]
    prefs['REPODIR'] = os.path.join(os.sep, *file_loc)

    # run any environment configuration commands
    env_results = {}
    for env_config, env_command in env.items():
        env_results[env_config] = call_series(env_command, env_config)

    # Create directory structure if needed
    #pdb.set_trace()
    for dir_name, dir_path in DIRECTORY_STRUCTURE.items():
        prefs[dir_name] = os.path.join(prefs['BASEDIR'], dir_path)
        make_dir(prefs[dir_name])

    # Create a launch script
    prefs_fn = os.path.join(prefs['BASEDIR'], 'prefs.json')
    launch_file = os.path.join(prefs['BASEDIR'], 'launch_autopilot.sh')

    if prefs['AGENT'] in ('PILOT', 'CHILD'):
        with open(launch_file, 'w') as launch_file_open:
            launch_file_open.write('#!/bin/bash\n')
            launch_file_open.write('killall jackd\n')
            launch_file_open.write('sudo killall pigpiod\n')
            launch_file_open.write('sudo mount -o remount,size=128M /dev/shm\n')
            if prefs['VENV']:
                launch_file_open.write("source " + os.path.join(prefs['VENV'], 'bin', 'activate')+'\n')
            launch_file_open.write('python3 -m autopilot.core.pilot -f {}'.format(prefs_fn))

    elif prefs['AGENT'] == 'TERMINAL':
        with open(launch_file, 'w') as launch_file_open:
            launch_file_open.write('#!/bin/bash\n')
            if prefs['VENV']:
                launch_file_open.write("source " + os.path.join(prefs['VENV'], 'bin', 'activate')+'\n')
            launch_file_open.write("python3 -m autopilot.core.terminal -f " + prefs_fn + "\n")


    config_msgs.append("Launch file created at {}".format(launch_file))
    os.chmod(launch_file, 0o775)

    # install as systemd service if requested
    if 'systemd' in env_params.keys():
        if env_params['systemd'] in (1, True):
            systemd_string = '''
[Unit]
Description=autopilot
After=multi-user.target

[Service]
Type=idle
ExecStart={launch_pi}

Restart=on-failure

[Install]
WantedBy=multi-user.target'''.format(launch_pi=launch_file)

            try:
                unit_loc = '/lib/systemd/system/autopilot.service'

                subprocess.call('sudo sh -c \"echo \'{}\' > {}\"'.format(systemd_string, unit_loc), shell=True)
                # enable the service
                subprocess.call(['sudo', 'systemctl', 'daemon-reload'])
                sysd_result = subprocess.call(['sudo', 'systemctl', 'enable', 'autopilot.service'])

                if sysd_result != 0:
                    error_msgs.append('Systemd service could not be enabled :(')
                else:
                    env_results['systemd'] = True
                    config_msgs.append('Systemd service installed and enabled, unit file written to {}'.format(unit_loc))

            except PermissionError:
                error_msgs.append("systemd service could not be installed due to a permissions error.\n"+\
                                  "create a unit file containing the following at {}\n\n{}".format(unit_loc, systemd_string))
                env_results['systemd'] = False



    ####################################
    # save prefs and finalize environment


    # save prefs
    #prefs_json = json.dumps(prefs, indent=4, separators=(',', ': '), sort_keys=True)
    #prefs_ret = subprocess.call('sudo sh -c \"echo \'{}\' > {}\"'.format(shlex.quote(prefs_json), shlex.quote(prefs_fn)), shell=True)
    # if prefs_ret != 0:
    #     error_msgs.append('Couldnt create prefs file :(')
    with open(prefs_fn, 'w') as prefs_f:
       json.dump(prefs, prefs_f, indent=4, separators=(',', ': '), sort_keys=True)

    # save basedir in autopilot user file
    with open(os.path.join(os.path.expanduser('~'), '.autopilot'), 'w') as autopilot_f:
        autopilot_f.write(prefs['BASEDIR'])





    #####################################3
    # User feedback

    env_result = "\033[0;32;40m\n--------------------------------\nEnvironment Configuration:\n"
    for config, result in env_results.items():
        if result:
            env_result += "  [ SUCCESS ] "
        else:
            env_result += "  [ FAILURE ] "

        env_result += config
        env_result += '\n'

    if venv_path:
        env_result += "  [ SUCCESS ] virtualenv detected, path: {}\n".format(venv_path)
    else:
        env_result += "  [ CMONDOG ] no virtualenv detected, running autopilot outside a venv is not recommended but it might work who knows\n"


    if len(config_msgs)>0:
        env_result += '\nAdditional Messages:'
        for msg in config_msgs:
            env_result += '  '
            env_result += msg
            env_result += '\n'



    env_result += '--------------------------------\u001b[0m'


    print('\n----------------------------------------')
    print('prefs.json has been created and saved to {}'.format(prefs_fn))
    pprint.pprint(prefs)
    print('----------------------------------------\n')

    print(env_result)

    if len(error_msgs)>0:
        for i, msg in enumerate(error_msgs):
            print('\033[1;37;41mSomething went wrong during setup, this is wrong thing #{}\u001b[0m'.format(i))
            print('\033[0;31;40m\n{}\n\u001b[0m'.format(msg))









    # TODO: After setup, create .autopilot file in user dir to point to autopilot dir

